#
# This file is part of openCARP
# (see https://www.openCARP.org).
#
# The openCARP project licenses this file to you under the 
# Apache License, Version 2.0 (the "License"); 
# you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""!
@file
Class for global bundle object.
It will contain all useful data and functions for creating a self-contained
bundle of the running experiment.

Use of the --bundle option:

    run.py --bundle exp_name

where exp_name is the name of the experiment.
If it is not specified, it is set to 'experiment' by default.

This option creates a self-contained bundle of the experiment which is run,
if the following constraints are followed:
in the run.py script, every path to an external file used in the experiment should
be encapsulated in the carputils.tools.simfile_path function.
Plus, if the file is a mesh file, the 'mesh' flag in the simfile_path function should
be set to True.

Note that additional files can be added to the bundle manually:
- if there is a folder named 'external' in run.py's directory, its content will be added
to the bundle automatically.
- if there is a folder named 'Mesh' in run.py's directory, its content will be added
to the bundle automatically.
- If the folder carputils.settings.config.MESH_DIR/exp_name/ exists, then its content will be added
to the bundle automatically.

Some metadata is also shared in the bundle: at the root of the bundle, a file named METADATA.yml
is created. This file is filled with metadata provided by the user in a file named METADATA.yml
(preferably using the template located at carputils/resources/METADATA.yml.tpl); it will also
contain autogenerated data about the environment in which the bundle was created:
- information about the creator found in carputils settings file
- the name given to the bundle
- the command line to use to run the experiment
- the python version used to create the bundle
"""

import sys
import os
from ruamel.yaml import YAML
from shutil import rmtree, copy2
from glob import glob
from pathlib import Path

from carputils import clean
from carputils import settings
from carputils import tools

from carputils.settings.config import yaml_format_dict
from carputils.bundle import bundleio
from carputils.bundle import parfileadjustment
from carputils.bundle import metadata


class BundleModule(object):

    def __init__(self):
        # Path to the protocol script
        self._exp_file = os.path.abspath(sys.modules['__main__'].__file__)
        # Name of the protocol script
        self._exp_filename = os.path.basename(self._exp_file)
        # Path of the experiment directory
        self._exp_dir = os.path.dirname(self._exp_file)
        # Name of mesh directory
        self._meshdir_name = 'Mesh'
        # Name of the external directory
        self._externaldir_name = 'external'
        # Name of the experiment
        self._exp_name = None
        # Name of the bundle
        self._bundle_foldername = None
        # Location of the bundle
        self._bundle_dir = None
        # String of protocol script
        self._protocol_script = None
        # OpenCARP parameters
        self._OCparams = None
        # OpenCARP parameters of type RFile, WFile
        self._file_OCparams = None
        # Number of calls of tools.simfile_path function
        self._nb_found_simfiles = 0
        # Lines at which tools.simfile_path function was used
        self._simfiles_lines = []
        # Information about simulation files given as an argument
        # to tools.simfile_path function.
        self._simfiles_queue = []

    @property
    def exp_file(self):
        """
        :return: Path to the protocol script
        """
        return self._exp_file

    @property
    def exp_filename(self):
        """
        :return: Name of the protocol script
        """
        return self._exp_filename

    @property
    def exp_dir(self):
        """
        :return: Directory containing the protocol script
        """
        return self._exp_dir

    @property
    def meshdir_name(self):
        """
        :return: Name of the mesh directory
        """
        return self._meshdir_name

    @property
    def externaldir_name(self):
        """
        :return: Name of the external directory
        """
        return self._externaldir_name

    @property
    def exp_name(self):
        """
        The name of the experiment, defined in the command line as an argument
        for the --bundle option.
        Default value: 'experiment'
        :return: Name of the experiment
        """
        if self._exp_name is None:
            self._exp_name = settings.cli.bundle
        return self._exp_name

    @property
    def bundle_foldername(self):
        """
        :return: Name of the bundle folder
        """
        if self._bundle_foldername is None:
            self._bundle_foldername = self.exp_name + '_bundle'
        return self._bundle_foldername

    @property
    def bundle_dir(self):
        """
        :return: Path to the bundle directory
        """
        if self._bundle_dir is None:
            self._bundle_dir = os.path.join(self.exp_dir, self.bundle_foldername)
        return self._bundle_dir

    @property
    def protocol_script(self):
        """
        :return: String containing the content of the protocol script
        """
        if self._protocol_script is None:
            self._protocol_script = bundleio.load_pscript(self.exp_file)
        return self._protocol_script

    @property
    def OCparams(self):
        """
        :return: List of OpenCARP parameters
        """
        if self._OCparams is None:
            self._OCparams = parfileadjustment.get_OCparams()
        return self._OCparams

    @property
    def file_OCparams(self):
        """
        :return: List of OpenCARP parameters giving path to a file
        (RFile and WFile parameters)
        """
        if self._file_OCparams is None:
            self._file_OCparams = parfileadjustment.filter_params(self.OCparams, ['RFile', 'WFile'])
        return self._file_OCparams

    def create_bundle_folder(self):
        """
        Creates the bundle folder, and its subfolders "mesh" and "external".
        Copies every file in 'mesh' or 'external' folders in it.
        :return: path to the bundle directory
        """
        # Check that the bundle folder doesn't already exists
        # and create it
        clean.overwrite_check(self.bundle_dir)
        os.mkdir(self.bundle_dir)

        # Copy local mesh folder to bundle
        bundleio.copy_folder_in_bundle(os.path.join(self.meshdir_name, self.exp_name),
                                       self.exp_dir, self.bundle_dir)
        # Copy MESH_DIR mesh folder in bundle
        bundleio.copy_mesh_dir_in_bundle(self.exp_name, self.bundle_dir)
        # Copy external folder in bundle
        os.mkdir(os.path.join(self.bundle_dir, self.externaldir_name))
        bundleio.copy_folder_in_bundle(self.externaldir_name, self.exp_dir, self.bundle_dir)
        return self.bundle_dir

    def finalize_bundle(self):
        """
        Writes the protocol script in the bundle folder,
        modifies parameter files containing unresolved paths
        and creates a zip archive of the bundle.
        Removes the bundle folder afterwards.
        """
        # Add queued external files to the bundle
        self.add_queued_simfiles()
        # Warn the user if required files are potentially missing
        self.check_found_simfiles()
        # Write protocol script in the bundle
        bundleio.write_pscript(self.protocol_script, self.exp_filename, self.bundle_dir)
        # Create metadata file in the bundle
        self.create_metadata_file()
        # Find all parameter files in bundle, adjust them and add missing files
        for folder, subfolders, filenames in os.walk(self.bundle_dir):
            for filename in filenames:
                if len(filename) > 4 and filename[-4:] == '.par':
                    filepath = os.path.join(folder, filename)
                    self.modify_par_file(filepath)
        print('Archiving bundle folder...')
        bundleio.create_zip(self.bundle_dir)
        print('Removing bundle folder...')
        rmtree(self.bundle_dir)

    def modify_par_file(self, par_file):
        """
        Modify parameter file in order for paths to be consistent in the bundle
        and put the input files given as parameters in the bundle.

        :param par_file: path to the parameter file to modify
        """
        f_src = open(par_file, 'r')
        content = f_src.read()
        comment_symbol = '#'
        affectation_symbol = '='

        # Read file lines
        content = content.splitlines()
        # String containing the modified file
        new_content = ''
        for line_content in content:
            mesh = False
            # Remove comments from the line
            line = line_content.split(comment_symbol, 1)[0]
            if line != '':
                if affectation_symbol in line:
                    # Get parameter value if it is a file parameter
                    split_param = line.split(affectation_symbol, 1)
                    args = split_param[1].split()
                    current_param = parfileadjustment.find_param(split_param[0], self.file_OCparams)
                    # Check if the parameter is in the string file parameters' list
                    # If it is, it needs a particular treatment
                    if current_param is None:
                        for sfp in parfileadjustment.STRING_FILEPARAMS:
                            if sfp.name in split_param[0]:
                                args = args[0].split(sfp.separator)
                                current_param = parfileadjustment.find_param(split_param[0], self.OCparams)
                                current_param._param_type = sfp.type
                                current_param._multiple_param = True
                    # Check if the parameter is a mesh path
                    if 'meshname' in split_param[0]:
                        mesh = True

                    # If the parameter is a file path, it is replaced and
                    # necessary files are copied
                    if mesh or current_param is not None:
                        for arg in args:
                            new_arg = self.resolve_parfile_path(arg, current_param, mesh)
                            line_content = line_content.replace(arg, new_arg, 1)
            new_content += line_content + '\n'
        f_src.close()

        # Overwrite the parameter file
        f_dest = open(par_file, 'w')
        f_dest.write(new_content)
        f_dest.close()

    def resolve_parfile_path(self, path, current_param=None, mesh=False):
        """
        Put the file targeted by 'path' into the bundle if necessary and returns
        the path to this file in the bundle.

        :param path: path to the file
        :param current_param: parameter (instance of _OCParam) associated to the path
        :param mesh: set to True if the file is a mesh file
        :return: the relative path to the file in the bundle
        """
        resolved_path = path
        is_path_abs = os.path.isabs(path)
        # Get absolute path to the file
        if not is_path_abs:
            path = os.path.join(self.exp_dir, path)
        path = os.path.normpath(path)

        # Case where the file is an output file.
        # Absolute paths are transformed to relative paths.
        if current_param is not None and current_param._param_type in ['WFile', 'WDir']:
            if is_path_abs:
                # If the file is in a subdirectory of the experiment,
                # it is transposed as a path relative to experiment directory
                if Path(self.exp_dir) in Path(path).parents:
                    resolved_path = os.path.relpath(path, self.exp_dir)
                # If the file is anywhere else, the path is changed
                # to the experiment directory
                else:
                    resolved_path = os.path.basename(path)

        # Case where the file is a read file
        # In this case, files have to be moved inside the bundle
        # and paths have to be changed
        elif mesh or (current_param is not None and current_param._param_type == 'RFile'):
            # Compute some useful names and paths, depending on the parameters
            if mesh:
                # File will be put in the bundle's mesh folder
                external_dir = os.path.join(self.exp_dir, self.meshdir_name)
                external_bundle_dir = os.path.join(self.bundle_dir, self.meshdir_name)
                externaldir_name = self.meshdir_name
            else:
                # File will be put in the bundle's external folder
                external_dir = os.path.join(self.exp_dir, self.externaldir_name)
                external_bundle_dir = os.path.join(self.bundle_dir, self.externaldir_name)
                externaldir_name = self.externaldir_name

            # Get the file(s) targeted by path parameter
            files = []
            if mesh:
                for ext in parfileadjustment.MESH_EXTENSIONS:
                    f = path + ext
                    if os.path.isfile(f):
                        files.append(f)
            else:
                if os.path.isfile(path):
                    files.append(path)
                if current_param._ext is not None:
                    f = path + current_param._ext
                    if os.path.isfile(f):
                        files.append(f)

            # The path is modified only if the file(s) exist
            if files:
                # Compute the relative path to the file in the bundle, and copy the
                # file at this location if needed.
                # Case where the file is in the MESH_DIR directory
                if mesh and Path(settings.config.MESH_DIR) in Path(path).parents:
                    resolved_path = os.path.join(externaldir_name,
                                                 os.path.relpath(path,
                                                                 settings.config.MESH_DIR))
                    # if the file is not in the self.exp_name directory, it has to be added in the bundle
                    if Path(os.path.join(settings.config.MESH_DIR, self.exp_name)) not in Path(path).parents:
                        for file in files:
                            dest_dir = os.path.join(external_bundle_dir,
                                                    os.path.dirname(os.path.relpath(file,
                                                                                    settings.config.MESH_DIR)))
                            os.makedirs(dest_dir, exist_ok=True)
                            copy2(file, dest_dir)

                # Case where the file is not at the right place for the bundle creation
                # and has to be added to the bundle
                elif Path(external_dir) not in Path(path).parents:
                    if Path(self.exp_dir) not in Path(path).parents:
                        resolved_path = os.path.join(externaldir_name, os.path.basename(path))
                        dest_dir = external_bundle_dir
                    else:
                        relpath = os.path.relpath(path, self.exp_dir)
                        resolved_path = os.path.join(externaldir_name, relpath)
                        dest_dir = os.path.join(external_bundle_dir, os.path.dirname(relpath))
                        os.makedirs(dest_dir, exist_ok=True)
                    for file in files:
                        copy2(file, dest_dir)

                # Case where the file is at the right place (and was already automatically put in the bundle)
                else:
                    resolved_path = os.path.join(externaldir_name, os.path.relpath(path, external_dir))

        return resolved_path

    def create_metadata_file(self):
        """
        Creates and writes the METADATA.yml file related to the current
        experiment in the bundle.
        """
        # Load the provided metadata file if it exists and merge it with the template
        metadata_obj = metadata.load_metadata_file(self.exp_dir)
        metadata_obj = metadata.autopopulate_template(metadata_obj)
        metadata_obj = metadata.add_env_metadata(metadata_obj)
        metadata_obj = metadata.add_exp_name(metadata_obj, self.exp_name)
        metadata_obj = metadata.add_command_line(metadata_obj, self.exp_filename)
        # Check if critical metadata is missing
        metadata.analyze_metadata(metadata_obj)
        # Write metadata file in the bundle
        with open(os.path.join(self.bundle_dir, metadata.METADATA_FILE), 'w') as fp:
            metadata.yaml.dump(metadata_obj, fp)

    def _add_found_simfile(self, lineno):
        """
        Should only be used in the tools.simfile_path function.
        Increment the number of simulation files found with the simfile_path function,
        and registers the corresponding line number in the protocol script.

        :param lineno: number of the line where the simfile_path function was used.
        """
        self._nb_found_simfiles += 1
        self._simfiles_lines.append(lineno)

    def _store_simfile(self, lineno, split_line, arg_index, path, mesh):
        """
        Should only be used in the tools.simfile_path function.
        This function tackles the case where tools.simfile_path is called before
        settings.cli is set (typically during the definition of the parser function).
        It stores the information concerning the file that has to be put in the bundle,
        so that the actual copy of the file and the modification of the protocol script
        can be postponed to the finalization of the bundle.

        :param lineno: number of the line where the simfile_path function was used.
        :param split_line: split of the line from where simfile_path was called.
        :param arg_index: index of the argument of simfile_path in split_line.
        :param path: the path given as an argument to simfile_path
        :param mesh: the value of 'mesh' parameter given to simfile_path
        """
        self._simfiles_queue.append((lineno, split_line, arg_index, path, mesh))

    def add_queued_simfiles(self):
        """
        Adds simfiles queued in self._simfiles_queue (with _store_simfile function)
        in the bundle and consequently modifies protocol script.
        """
        while self._simfiles_queue:
            lineno, split_line, arg_index, path, mesh = self._simfiles_queue.pop()
            # Get the string that must replace the function parameter.
            resolved_path = tools.resolve_simfilepath(path, mesh)
            # Replace path by resolved path in function call
            split_line[arg_index] = resolved_path
            split_script = self.protocol_script.splitlines()
            split_script[lineno - 1] = ''.join(split_line)
            # Update protocol script
            self._protocol_script = '\n'.join(split_script)

    def check_found_simfiles(self):
        """
        Checks if any external files could be missing in the bundle for the
        experiment to run properly and raises adequate warnings.
        """
        split_script = self.protocol_script.splitlines()
        if self._nb_found_simfiles == 0:
            print("WARNING: the tools.simfile_path function wasn't called during the creation of the bundle.\n",
                  "Some required files could be missing in the bundle.\n")
        for lineno, line in enumerate(split_script):
            if 'simfile_path(' in line and lineno+1 not in self._simfiles_lines:
                print("WARNING: file referred to at line {} may not have been added to the bundle.\n".format(lineno),
                      "In order for this file to be added, it has to be in the 'external' or 'Mesh' folder.\n",
                      "Context:\n",
                      line, "\n")
        for lineno in self._simfiles_lines:
            if self._simfiles_lines.count(lineno) > 1:
                print("WARNING: line {} seems to have been executed multiple times.\n".format(lineno),
                      "It could not behave as expected in the bundle.\n",
                      "Context:\n",
                      split_script[lineno+1], "\n")
