#
# This file is part of openCARP
# (see https://www.openCARP.org).
#
# The openCARP project licenses this file to you under the
# Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from .general import AbstractIonicModel as _AbstractIonicModel
from collections import OrderedDict as _OrderedDict
from carputils.format import stringtable


class PassiveIonicRegion(_AbstractIonicModel):
    """
    Describes the passive ionic model configuration

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
    """
    SHORTNAME = 'passive'
    MODELID = 'PASSIVE'
    PARAMETERS = {}


class AlievPanfilovIonicModel(_AbstractIonicModel):
    """
    Describes the AlievPanfilov ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        mu1 : float, optional
            Value of mu1, defaults to 0.2
        mu2 : float, optional
            Value of mu2, defaults to 0.3
    """
    SHORTNAME = 'alievpanfilov'
    MODELID = 'AlievPanfilov'
    PARAMETERS = {'mu1': 0.2,  'mu2': 0.3}


class ARPFIonicModel(_AbstractIonicModel):
    """
    Describes the ARPF ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_e : float, optional
            Value of Ca_e, defaults to 1.8
        Ca_handling : float, optional
            Value of Ca_handling, defaults to 1.0
        Cl_e : float, optional
            Value of Cl_e, defaults to 150.0
        Cl_i : float, optional
            Value of Cl_i, defaults to 30.0
        g_CaB : float, optional
            Value of g_CaB, defaults to 0.00035182
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.27
        g_CaT : float, optional
            Value of g_CaT, defaults to 0.2
        g_Cl : float, optional
            Value of g_Cl, defaults to 0.3
        g_ClB : float, optional
            Value of g_ClB, defaults to 0.0009
        g_K1 : float, optional
            Value of g_K1, defaults to 0.5
        g_KB : float, optional
            Value of g_KB, defaults to 5e-05
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.015583333333333
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.011738183745949
        g_Na : float, optional
            Value of g_Na, defaults to 20.0
        g_NaB : float, optional
            Value of g_NaB, defaults to 2.97e-05
        g_NaL : float, optional
            Value of g_NaL, defaults to 0.0162
        g_to : float, optional
            Value of g_to, defaults to 0.112
        I_CaP_Hill : float, optional
            Value of I_CaP_Hill, defaults to 1.6
        I_CaP_Kmf : float, optional
            Value of I_CaP_Kmf, defaults to 0.5
        I_CaP_Vmf : float, optional
            Value of I_CaP_Vmf
        inacamax : float, optional
            Value of inacamax, defaults to 4.5
        K_e_init : float, optional
            Value of K_e_init, defaults to 5.4
        K_i : float, optional
            Value of K_i, defaults to 135.0
        kmcaact : float, optional
            Value of kmcaact, defaults to 0.000125
        kmcai : float, optional
            Value of kmcai, defaults to 0.0036
        kmcao : float, optional
            Value of kmcao, defaults to 1.3
        kmnai1 : float, optional
            Value of kmnai1, defaults to 12.3
        kmnao : float, optional
            Value of kmnao, defaults to 87.5
        ksat : float, optional
            Value of ksat, defaults to 0.27
        Mg_i : float, optional
            Value of Mg_i, defaults to 1.0
        Na_e : float, optional
            Value of Na_e, defaults to 140.0
        Na_i : float, optional
            Value of Na_i, defaults to 8.8
        nu : float, optional
            Value of nu, defaults to 0.35
        P_NaK : float, optional
            Value of P_NaK, defaults to 0.01833
        T : float, optional
            Value of T, defaults to 310.0
    """
    SHORTNAME = 'arpf'
    MODELID = 'ARPF'
    PARAMETERS = {'Ca_e': 1.8,  'Ca_handling': 1.0,  'Cl_e': 150.0,  
                  'Cl_i': 30.0,  'g_CaB': 0.00035182,  'g_CaL': 0.27,  
                  'g_CaT': 0.2,  'g_Cl': 0.3,  'g_ClB': 0.0009,  'g_K1': 0.5,  
                  'g_KB': 5e-05,  'g_Kr': 0.015583333333333,  
                  'g_Ks': 0.011738183745949,  'g_Na': 20.0,  'g_NaB': 2.97e-05,  
                  'g_NaL': 0.0162,  'g_to': 0.112,  'I_CaP_Hill': 1.6,  
                  'I_CaP_Kmf': 0.5,  'I_CaP_Vmf': None,  'inacamax': 4.5,  
                  'K_e_init': 5.4,  'K_i': 135.0,  'kmcaact': 0.000125,  
                  'kmcai': 0.0036,  'kmcao': 1.3,  'kmnai1': 12.3,  
                  'kmnao': 87.5,  'ksat': 0.27,  'Mg_i': 1.0,  'Na_e': 140.0,  
                  'Na_i': 8.8,  'nu': 0.35,  'P_NaK': 0.01833,  'T': 310.0}


class StewardIonicModel(_AbstractIonicModel):
    """
    Describes the Steward ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        g_bca : float, optional
            Value of g_bca, defaults to 0.000592
        g_bna : float, optional
            Value of g_bna, defaults to 0.00029
        g_CaL : float, optional
            Value of g_CaL, defaults to 3.98e-05
        g_f_K : float, optional
            Value of g_f_K, defaults to 0.0234346
        g_f_Na : float, optional
            Value of g_f_Na, defaults to 0.0145654
        g_K1 : float, optional
            Value of g_K1, defaults to 0.065
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0918
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.2352
        g_Na : float, optional
            Value of g_Na, defaults to 130.5744
        g_pCa : float, optional
            Value of g_pCa, defaults to 0.1238
        g_pK : float, optional
            Value of g_pK, defaults to 0.0146
        g_sus : float, optional
            Value of g_sus, defaults to 0.0227
        g_to : float, optional
            Value of g_to, defaults to 0.08184
        K_NaCa : float, optional
            Value of K_NaCa, defaults to 1000.0
        P_NaK : float, optional
            Value of P_NaK, defaults to 2.724
    """
    SHORTNAME = 'steward'
    MODELID = 'Steward'
    PARAMETERS = {'g_bca': 0.000592,  'g_bna': 0.00029,  'g_CaL': 3.98e-05,  
                  'g_f_K': 0.0234346,  'g_f_Na': 0.0145654,  'g_K1': 0.065,  
                  'g_Kr': 0.0918,  'g_Ks': 0.2352,  'g_Na': 130.5744,  
                  'g_pCa': 0.1238,  'g_pK': 0.0146,  'g_sus': 0.0227,  
                  'g_to': 0.08184,  'K_NaCa': 1000.0,  'P_NaK': 2.724}


class PRd2IonicModel(_AbstractIonicModel):
    """
    Describes the PRd2 ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        gRyR3 : float, optional
            Value of gRyR3, defaults to 15.0
        RyR2_sens : float, optional
            Value of RyR2_sens, defaults to 1.0
        RyR2_tau : float, optional
            Value of RyR2_tau, defaults to 6.0
    """
    SHORTNAME = 'prd2'
    MODELID = 'PRd2'
    PARAMETERS = {'gRyR3': 15.0,  'RyR2_sens': 1.0,  'RyR2_tau': 6.0}


class converted_COURTEMANCHEIonicModel(_AbstractIonicModel):
    """
    Describes the converted_COURTEMANCHE ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        blf_i_Kur : float, optional
            Value of blf_i_Kur, defaults to 1.0
        blf_i_up : float, optional
            Value of blf_i_up, defaults to 1.0
        Ca_o : float, optional
            Value of Ca_o, defaults to 1.8
        g_bCa : float, optional
            Value of g_bCa, defaults to 0.00113
        g_bNa : float, optional
            Value of g_bNa, defaults to 0.000674
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.1238
        g_K1 : float, optional
            Value of g_K1, defaults to 0.09
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0294
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.129
        g_Na : float, optional
            Value of g_Na, defaults to 7.8
        g_to : float, optional
            Value of g_to, defaults to 0.1652
        K_o : float, optional
            Value of K_o, defaults to 5.4
        Na_o : float, optional
            Value of Na_o, defaults to 140.0
    """
    SHORTNAME = 'converted_courtemanche'
    MODELID = 'converted_COURTEMANCHE'
    PARAMETERS = {'blf_i_Kur': 1.0,  'blf_i_up': 1.0,  'Ca_o': 1.8,  
                  'g_bCa': 0.00113,  'g_bNa': 0.000674,  'g_CaL': 0.1238,  
                  'g_K1': 0.09,  'g_Kr': 0.0294,  'g_Ks': 0.129,  'g_Na': 7.8,  
                  'g_to': 0.1652,  'K_o': 5.4,  'Na_o': 140.0}


class converted_DiFranNobleIonicModel(_AbstractIonicModel):
    """
    Describes the converted_DiFranNoble ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        C : float, optional
            Value of C
        G_Na : float, optional
            Value of G_Na
        kNaCa : float, optional
            Value of kNaCa, defaults to 0.02
        p_to : float, optional
            Value of p_to, defaults to 1.0
        pCaf : float, optional
            Value of pCaf, defaults to 1.0
        pFunny : float, optional
            Value of pFunny, defaults to 1.0
        pK : float, optional
            Value of pK, defaults to 1.0
    """
    SHORTNAME = 'converted_difrannoble'
    MODELID = 'converted_DiFranNoble'
    PARAMETERS = {'C': None,  'G_Na': None,  'kNaCa': 0.02,  'p_to': 1.0,  
                  'pCaf': 1.0,  'pFunny': 1.0,  'pK': 1.0}


class converted_LRDII_FIonicModel(_AbstractIonicModel):
    """
    Describes the converted_LRDII_F ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        g_NaK : float, optional
            Value of g_NaK, defaults to 1.0
    """
    SHORTNAME = 'converted_lrdii_f'
    MODELID = 'converted_LRDII_F'
    PARAMETERS = {'g_NaK': 1.0}


class converted_MBRDRIonicModel(_AbstractIonicModel):
    """
    Describes the converted_MBRDR ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        APDshorten : float, optional
            Value of APDshorten, defaults to 1.0
        GNa : float, optional
            Value of GNa, defaults to 15.0
        Gsi : float, optional
            Value of Gsi, defaults to 0.09
    """
    SHORTNAME = 'converted_mbrdr'
    MODELID = 'converted_MBRDR'
    PARAMETERS = {'APDshorten': 1.0,  'GNa': 15.0,  'Gsi': 0.09}


class converted_NYGRENIonicModel(_AbstractIonicModel):
    """
    Describes the converted_NYGREN ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        G_CaB : float, optional
            Value of G_CaB, defaults to 0.00157362
        G_CaL : float, optional
            Value of G_CaL, defaults to 0.135
        G_K1 : float, optional
            Value of G_K1, defaults to 0.06
        G_Kr : float, optional
            Value of G_Kr, defaults to 0.01
        G_KS : float, optional
            Value of G_KS, defaults to 0.02
        G_NaB : float, optional
            Value of G_NaB, defaults to 0.00121198
        G_sus : float, optional
            Value of G_sus, defaults to 0.055
        G_t : float, optional
            Value of G_t, defaults to 0.15
        I_NaKmax : float, optional
            Value of I_NaKmax, defaults to 1.416506
        k_NaCa : float, optional
            Value of k_NaCa, defaults to 0.000749684
    """
    SHORTNAME = 'converted_nygren'
    MODELID = 'converted_NYGREN'
    PARAMETERS = {'G_CaB': 0.00157362,  'G_CaL': 0.135,  'G_K1': 0.06,  
                  'G_Kr': 0.01,  'G_KS': 0.02,  'G_NaB': 0.00121198,  
                  'G_sus': 0.055,  'G_t': 0.15,  'I_NaKmax': 1.416506,  
                  'k_NaCa': 0.000749684}


class converted_RNCIonicModel(_AbstractIonicModel):
    """
    Describes the converted_RNC ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_o : float, optional
            Value of Ca_o, defaults to 1.8
        Cl_o : float, optional
            Value of Cl_o, defaults to 132.0
        g_bCa : float, optional
            Value of g_bCa, defaults to 0.00113
        g_bNa : float, optional
            Value of g_bNa, defaults to 0.000674
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.24
        g_ClCa : float, optional
            Value of g_ClCa, defaults to 0.3
        g_K1 : float, optional
            Value of g_K1, defaults to 0.15
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.06984
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.0561
        g_Na : float, optional
            Value of g_Na, defaults to 7.8
        g_to : float, optional
            Value of g_to, defaults to 0.19824
        K_o : float, optional
            Value of K_o, defaults to 5.4
        maxCa_up : float, optional
            Value of maxCa_up, defaults to 15.0
        maxI_NaCa : float, optional
            Value of maxI_NaCa, defaults to 1600.0
        maxI_NaK : float, optional
            Value of maxI_NaK, defaults to 0.6
        maxI_pCa : float, optional
            Value of maxI_pCa, defaults to 0.275
        maxI_up : float, optional
            Value of maxI_up, defaults to 0.005
        Na_o : float, optional
            Value of Na_o, defaults to 140.0
        tau_tr : float, optional
            Value of tau_tr, defaults to 180.0
        tau_u : float, optional
            Value of tau_u, defaults to 8.0
    """
    SHORTNAME = 'converted_rnc'
    MODELID = 'converted_RNC'
    PARAMETERS = {'Ca_o': 1.8,  'Cl_o': 132.0,  'g_bCa': 0.00113,  
                  'g_bNa': 0.000674,  'g_CaL': 0.24,  'g_ClCa': 0.3,  
                  'g_K1': 0.15,  'g_Kr': 0.06984,  'g_Ks': 0.0561,  
                  'g_Na': 7.8,  'g_to': 0.19824,  'K_o': 5.4,  
                  'maxCa_up': 15.0,  'maxI_NaCa': 1600.0,  'maxI_NaK': 0.6,  
                  'maxI_pCa': 0.275,  'maxI_up': 0.005,  'Na_o': 140.0,  
                  'tau_tr': 180.0,  'tau_u': 8.0}


class converted_TT2IonicModel(_AbstractIonicModel):
    """
    Describes the converted_TT2 ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Bufc : float, optional
            Value of Bufc, defaults to 0.2
        Bufsr : float, optional
            Value of Bufsr, defaults to 10.0
        Bufss : float, optional
            Value of Bufss, defaults to 0.4
        Cao : float, optional
            Value of Cao, defaults to 2.0
        CAPACITANCE : float, optional
            Value of CAPACITANCE, defaults to 0.185
        cell_type : float, optional
            Value of cell_type
        D_CaL_off : float, optional
            Value of D_CaL_off, defaults to 0.0
        Fconst : float, optional
            Value of Fconst, defaults to 96485.3415
        flags : float, optional
            Value of flags
        GbCa : float, optional
            Value of GbCa, defaults to 0.000592
        GbNa : float, optional
            Value of GbNa, defaults to 0.00029
        GCaL : float, optional
            Value of GCaL, defaults to 3.98e-05
        GK1 : float, optional
            Value of GK1, defaults to 5.405
        Gkr : float, optional
            Value of Gkr, defaults to 0.153
        Gks : float, optional
            Value of Gks
        GNa : float, optional
            Value of GNa, defaults to 14.838
        GpCa : float, optional
            Value of GpCa, defaults to 0.1238
        GpK : float, optional
            Value of GpK, defaults to 0.0146
        Gto : float, optional
            Value of Gto
        Kbufc : float, optional
            Value of Kbufc, defaults to 0.001
        Kbufsr : float, optional
            Value of Kbufsr, defaults to 0.3
        Kbufss : float, optional
            Value of Kbufss, defaults to 0.00025
        KmCa : float, optional
            Value of KmCa, defaults to 1.38
        KmK : float, optional
            Value of KmK, defaults to 1.0
        KmNa : float, optional
            Value of KmNa, defaults to 40.0
        KmNai : float, optional
            Value of KmNai, defaults to 87.5
        knaca : float, optional
            Value of knaca, defaults to 1000.0
        knak : float, optional
            Value of knak, defaults to 2.724
        Ko : float, optional
            Value of Ko, defaults to 5.4
        KpCa : float, optional
            Value of KpCa, defaults to 0.0005
        ksat : float, optional
            Value of ksat, defaults to 0.1
        Kup : float, optional
            Value of Kup, defaults to 0.00025
        n : float, optional
            Value of n, defaults to 0.35
        Nao : float, optional
            Value of Nao, defaults to 140.0
        pKNa : float, optional
            Value of pKNa, defaults to 0.03
        Rconst : float, optional
            Value of Rconst, defaults to 8314.472
        scl_tau_f : float, optional
            Value of scl_tau_f, defaults to 1.0
        T : float, optional
            Value of T, defaults to 310.0
        Vc : float, optional
            Value of Vc, defaults to 0.016404
        Vleak : float, optional
            Value of Vleak, defaults to 0.00036
        Vmaxup : float, optional
            Value of Vmaxup, defaults to 0.006375
        Vrel : float, optional
            Value of Vrel, defaults to 0.102
        Vsr : float, optional
            Value of Vsr, defaults to 0.001094
        Vss : float, optional
            Value of Vss, defaults to 5.468e-05
        xr2_off : float, optional
            Value of xr2_off, defaults to 0.0
    """
    SHORTNAME = 'converted_tt2'
    MODELID = 'converted_TT2'
    PARAMETERS = {'Bufc': 0.2,  'Bufsr': 10.0,  'Bufss': 0.4,  'Cao': 2.0,  
                  'CAPACITANCE': 0.185,  'cell_type': None,  'D_CaL_off': 0.0,  
                  'Fconst': 96485.3415,  'flags': None,  'GbCa': 0.000592,  
                  'GbNa': 0.00029,  'GCaL': 3.98e-05,  'GK1': 5.405,  
                  'Gkr': 0.153,  'Gks': None,  'GNa': 14.838,  'GpCa': 0.1238,  
                  'GpK': 0.0146,  'Gto': None,  'Kbufc': 0.001,  'Kbufsr': 0.3,  
                  'Kbufss': 0.00025,  'KmCa': 1.38,  'KmK': 1.0,  'KmNa': 40.0,  
                  'KmNai': 87.5,  'knaca': 1000.0,  'knak': 2.724,  'Ko': 5.4,  
                  'KpCa': 0.0005,  'ksat': 0.1,  'Kup': 0.00025,  'n': 0.35,  
                  'Nao': 140.0,  'pKNa': 0.03,  'Rconst': 8314.472,  
                  'scl_tau_f': 1.0,  'T': 310.0,  'Vc': 0.016404,  
                  'Vleak': 0.00036,  'Vmaxup': 0.006375,  'Vrel': 0.102,  
                  'Vsr': 0.001094,  'Vss': 5.468e-05,  'xr2_off': 0.0}


class converted_UCLA_RABIonicModel(_AbstractIonicModel):
    """
    Describes the converted_UCLA_RAB ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        BASE_Ki : float, optional
            Value of BASE_Ki, defaults to 140.0
        blf_ICaL : float, optional
            Value of blf_ICaL, defaults to 1.0
        blf_IKr : float, optional
            Value of blf_IKr, defaults to 1.0
        blf_IKs : float, optional
            Value of blf_IKs, defaults to 1.0
        blf_INa : float, optional
            Value of blf_INa, defaults to 1.0
        blf_INaCa : float, optional
            Value of blf_INaCa, defaults to 1.0
        Cao : float, optional
            Value of Cao
        constant_K_i : float, optional
            Value of constant_K_i, defaults to 0.0
        gK1 : float, optional
            Value of gK1, defaults to 0.3
        gKr : float, optional
            Value of gKr, defaults to 0.0125
        gKs : float, optional
            Value of gKs
        gNa : float, optional
            Value of gNa, defaults to 12.0
        gss : float, optional
            Value of gss
        gtof : float, optional
            Value of gtof, defaults to 0.11
        gtos : float, optional
            Value of gtos, defaults to 0.04
        Ko : float, optional
            Value of Ko
        m : float, optional
            Value of m
        Na_i : float, optional
            Value of Na_i
        Nao : float, optional
            Value of Nao
        T : float, optional
            Value of T, defaults to 308.0
    """
    SHORTNAME = 'converted_ucla_rab'
    MODELID = 'converted_UCLA_RAB'
    PARAMETERS = {'BASE_Ki': 140.0,  'blf_ICaL': 1.0,  'blf_IKr': 1.0,  
                  'blf_IKs': 1.0,  'blf_INa': 1.0,  'blf_INaCa': 1.0,  
                  'Cao': None,  'constant_K_i': 0.0,  'gK1': 0.3,  
                  'gKr': 0.0125,  'gKs': None,  'gNa': 12.0,  'gss': None,  
                  'gtof': 0.11,  'gtos': 0.04,  'Ko': None,  'm': None,  
                  'Na_i': None,  'Nao': None,  'T': 308.0}


class UCLA_RAB_SCRIonicModel(_AbstractIonicModel):
    """
    Describes the UCLA_RAB_SCR ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_o : float, optional
            Value of Ca_o, defaults to 1.8
        CSR : float, optional
            Value of CSR, defaults to 1200.0
        g_Ca : float, optional
            Value of g_Ca, defaults to 0.6
        g_K1 : float, optional
            Value of g_K1, defaults to 0.3
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.003
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.043
        g_Na : float, optional
            Value of g_Na, defaults to 12.0
        g_RyR : float, optional
            Value of g_RyR, defaults to 2.25
        g_sp : float, optional
            Value of g_sp
        g_tof : float, optional
            Value of g_tof, defaults to 0.1
        g_tos : float, optional
            Value of g_tos, defaults to 0.04
        i_NaCamax : float, optional
            Value of i_NaCamax, defaults to 0.3
        i_NaKmax : float, optional
            Value of i_NaKmax, defaults to 1.5
        K_o : float, optional
            Value of K_o, defaults to 5.4
        Na_o : float, optional
            Value of Na_o, defaults to 136.0
        rseed : float, optional
            Value of rseed, defaults to -1.0
        t_SCR : float, optional
            Value of t_SCR, defaults to -1.0
        v_Na : float, optional
            Value of v_Na, defaults to 0.0
        v_up : float, optional
            Value of v_up, defaults to 0.3
    """
    SHORTNAME = 'ucla_rab_scr'
    MODELID = 'UCLA_RAB_SCR'
    PARAMETERS = {'Ca_o': 1.8,  'CSR': 1200.0,  'g_Ca': 0.6,  'g_K1': 0.3,  
                  'g_Kr': 0.003,  'g_Ks': 0.043,  'g_Na': 12.0,  'g_RyR': 2.25,  
                  'g_sp': None,  'g_tof': 0.1,  'g_tos': 0.04,  
                  'i_NaCamax': 0.3,  'i_NaKmax': 1.5,  'K_o': 5.4,  
                  'Na_o': 136.0,  'rseed': -1.0,  't_SCR': -1.0,  'v_Na': 0.0,  
                  'v_up': 0.3}


class COURTEMANCHEIonicModel(_AbstractIonicModel):
    """
    Describes the COURTEMANCHE ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        blf_g_Kur : float, optional
            Value of blf_g_Kur, defaults to 1.0
        extern_Cab : float, optional
            Value of extern_Cab
        g_bCa : float, optional
            Value of g_bCa, defaults to 0.00113
        g_bNa : float, optional
            Value of g_bNa, defaults to 0.000674
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.1238
        g_K1 : float, optional
            Value of g_K1, defaults to 0.09
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0294
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.129
        g_Na : float, optional
            Value of g_Na, defaults to 7.8
        g_to : float, optional
            Value of g_to, defaults to 0.1652
    """
    SHORTNAME = 'courtemanche'
    MODELID = 'COURTEMANCHE'
    PARAMETERS = {'blf_g_Kur': 1.0,  'extern_Cab': None,  'g_bCa': 0.00113,  
                  'g_bNa': 0.000674,  'g_CaL': 0.1238,  'g_K1': 0.09,  
                  'g_Kr': 0.0294,  'g_Ks': 0.129,  'g_Na': 7.8,  
                  'g_to': 0.1652}


class DiFranNobleIonicModel(_AbstractIonicModel):
    """
    Describes the DiFranNoble ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        C : float, optional
            Value of C
        G_Na : float, optional
            Value of G_Na, defaults to 750.0
        kNaCa : float, optional
            Value of kNaCa, defaults to 0.02
        p_to : float, optional
            Value of p_to, defaults to 1.0
        pCaf : float, optional
            Value of pCaf, defaults to 1.0
        pFunny : float, optional
            Value of pFunny, defaults to 1.0
        pK : float, optional
            Value of pK, defaults to 1.0
    """
    SHORTNAME = 'difrannoble'
    MODELID = 'DiFranNoble'
    PARAMETERS = {'C': None,  'G_Na': 750.0,  'kNaCa': 0.02,  'p_to': 1.0,  
                  'pCaf': 1.0,  'pFunny': 1.0,  'pK': 1.0}


class ECMEIonicModel(_AbstractIonicModel):
    """
    Describes the ECME ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Cao : float, optional
            Value of Cao, defaults to 2.0
        etSOD : float, optional
            Value of etSOD, defaults to 0.00143
        gNa : float, optional
            Value of gNa, defaults to 12.8
        H_KATP : float, optional
            Value of H_KATP, defaults to -0.001
        INaKmax : float, optional
            Value of INaKmax, defaults to 3.147
        j_Vt2SO2m : float, optional
            Value of j_Vt2SO2m, defaults to 0.1
        Ko : float, optional
            Value of Ko, defaults to 5.4
        Nao : float, optional
            Value of Nao, defaults to 140.0
        rhoF1 : float, optional
            Value of rhoF1, defaults to 0.05
        rhoREF : float, optional
            Value of rhoREF, defaults to 0.000375
        rhoREN : float, optional
            Value of rhoREN, defaults to 0.0001
        shunt : float, optional
            Value of shunt, defaults to 0.1
        sigma_KATP : float, optional
            Value of sigma_KATP, defaults to 0.6
        useCK : float, optional
            Value of useCK
        Vmuni : float, optional
            Value of Vmuni, defaults to 0.0275
    """
    SHORTNAME = 'ecme'
    MODELID = 'ECME'
    PARAMETERS = {'Cao': 2.0,  'etSOD': 0.00143,  'gNa': 12.8,  
                  'H_KATP': -0.001,  'INaKmax': 3.147,  'j_Vt2SO2m': 0.1,  
                  'Ko': 5.4,  'Nao': 140.0,  'rhoF1': 0.05,  
                  'rhoREF': 0.000375,  'rhoREN': 0.0001,  'shunt': 0.1,  
                  'sigma_KATP': 0.6,  'useCK': None,  'Vmuni': 0.0275}


class FHN_RMcCIonicModel(_AbstractIonicModel):
    """
    Describes the FHN_RMcC ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        eta1 : float, optional
            Value of eta1, defaults to 4.4
        eta2 : float, optional
            Value of eta2, defaults to 0.012
        eta3 : float, optional
            Value of eta3, defaults to 1.0
        G : float, optional
            Value of G, defaults to 1.5
        vp : float, optional
            Value of vp, defaults to 100.0
        vth : float, optional
            Value of vth, defaults to 13.0
    """
    SHORTNAME = 'fhn_rmcc'
    MODELID = 'FHN_RMcC'
    PARAMETERS = {'eta1': 4.4,  'eta2': 0.012,  'eta3': 1.0,  'G': 1.5,  
                  'vp': 100.0,  'vth': 13.0}


class FOXIonicModel(_AbstractIonicModel):
    """
    Describes the FOX ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0136
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.0245
        g_Na : float, optional
            Value of g_Na, defaults to 12.8
        P_Ca : float, optional
            Value of P_Ca, defaults to 2.26e-05
    """
    SHORTNAME = 'fox'
    MODELID = 'FOX'
    PARAMETERS = {'g_Kr': 0.0136,  'g_Ks': 0.0245,  'g_Na': 12.8,  
                  'P_Ca': 2.26e-05}


class GPBIonicModel(_AbstractIonicModel):
    """
    Describes the GPB ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.08597401
        Cao : float, optional
            Value of Cao, defaults to 1.8
        cell_type : float, optional
            Value of cell_type
        Cli : float, optional
            Value of Cli, defaults to 15.0
        Clo : float, optional
            Value of Clo, defaults to 150.0
        flags : float, optional
            Value of flags
        Ki_init : float, optional
            Value of Ki_init, defaults to 120.0
        Ko : float, optional
            Value of Ko, defaults to 5.4
        markov_iks : float, optional
            Value of markov_iks, defaults to 0.0
        Mgi : float, optional
            Value of Mgi, defaults to 1.0
        Nai_init : float, optional
            Value of Nai_init, defaults to 9.06
        Nao : float, optional
            Value of Nao, defaults to 140.0
    """
    SHORTNAME = 'gpb'
    MODELID = 'GPB'
    PARAMETERS = {'Ca_i_init': 0.08597401,  'Cao': 1.8,  'cell_type': None,  
                  'Cli': 15.0,  'Clo': 150.0,  'flags': None,  
                  'Ki_init': 120.0,  'Ko': 5.4,  'markov_iks': 0.0,  
                  'Mgi': 1.0,  'Nai_init': 9.06,  'Nao': 140.0}


class GPVatriaIonicModel(_AbstractIonicModel):
    """
    Describes the GPVatria ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        AF : float, optional
            Value of AF, defaults to 0.0
        factor_g_CaL : float, optional
            Value of factor_g_CaL, defaults to 1.0
        factor_g_K1 : float, optional
            Value of factor_g_K1, defaults to 1.0
        factor_J_SRCarel : float, optional
            Value of factor_J_SRCarel, defaults to 1.0
        factor_J_SRleak : float, optional
            Value of factor_J_SRleak, defaults to 1.0
        factor_Vmax_SRCaP : float, optional
            Value of factor_Vmax_SRCaP, defaults to 1.0
        ISO : float, optional
            Value of ISO, defaults to 0.0
        markov_iks : float, optional
            Value of markov_iks, defaults to 0.0
        RA : float, optional
            Value of RA, defaults to 0.0
        usefca : float, optional
            Value of usefca, defaults to 0.0
    """
    SHORTNAME = 'gpvatria'
    MODELID = 'GPVatria'
    PARAMETERS = {'AF': 0.0,  'factor_g_CaL': 1.0,  'factor_g_K1': 1.0,  
                  'factor_J_SRCarel': 1.0,  'factor_J_SRleak': 1.0,  
                  'factor_Vmax_SRCaP': 1.0,  'ISO': 0.0,  'markov_iks': 0.0,  
                  'RA': 0.0,  'usefca': 0.0}


class GPB_LandIonicModel(_AbstractIonicModel):
    """
    Describes the GPB_Land ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        a : float, optional
            Value of a, defaults to 0.35
        A_1 : float, optional
            Value of A_1, defaults to -29.0
        A_2 : float, optional
            Value of A_2, defaults to 116.0
        alpha_1 : float, optional
            Value of alpha_1, defaults to 0.1
        alpha_2 : float, optional
            Value of alpha_2, defaults to 0.5
        beta_0 : float, optional
            Value of beta_0, defaults to 1.65
        beta_1 : float, optional
            Value of beta_1, defaults to -1.5
        Ca_50ref : float, optional
            Value of Ca_50ref, defaults to 0.52
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.08597401
        Cao : float, optional
            Value of Cao, defaults to 1.8
        cell_type : float, optional
            Value of cell_type
        Cli : float, optional
            Value of Cli, defaults to 15.0
        Clo : float, optional
            Value of Clo, defaults to 150.0
        flags : float, optional
            Value of flags
        k_TRPN : float, optional
            Value of k_TRPN, defaults to 0.14
        k_xb : float, optional
            Value of k_xb, defaults to 0.0049
        Ki_init : float, optional
            Value of Ki_init, defaults to 120.0
        Ko : float, optional
            Value of Ko, defaults to 5.4
        lengthDep : float, optional
            Value of lengthDep, defaults to 0.0
        markov_iks : float, optional
            Value of markov_iks, defaults to 0.0
        Mgi : float, optional
            Value of Mgi, defaults to 1.0
        n_TRPN : float, optional
            Value of n_TRPN, defaults to 1.54
        n_xb : float, optional
            Value of n_xb, defaults to 3.38
        Nai_init : float, optional
            Value of Nai_init, defaults to 9.06
        Nao : float, optional
            Value of Nao, defaults to 140.0
        T_ref : float, optional
            Value of T_ref, defaults to 117.1
        TRPN_50 : float, optional
            Value of TRPN_50, defaults to 0.37
    """
    SHORTNAME = 'gpb_land'
    MODELID = 'GPB_Land'
    PARAMETERS = {'a': 0.35,  'A_1': -29.0,  'A_2': 116.0,  'alpha_1': 0.1,  
                  'alpha_2': 0.5,  'beta_0': 1.65,  'beta_1': -1.5,  
                  'Ca_50ref': 0.52,  'Ca_i_init': 0.08597401,  'Cao': 1.8,  
                  'cell_type': None,  'Cli': 15.0,  'Clo': 150.0,  
                  'flags': None,  'k_TRPN': 0.14,  'k_xb': 0.0049,  
                  'Ki_init': 120.0,  'Ko': 5.4,  'lengthDep': 0.0,  
                  'markov_iks': 0.0,  'Mgi': 1.0,  'n_TRPN': 1.54,  
                  'n_xb': 3.38,  'Nai_init': 9.06,  'Nao': 140.0,  
                  'T_ref': 117.1,  'TRPN_50': 0.37}


class GPB_LandHumanIonicModel(_AbstractIonicModel):
    """
    Describes the GPB_LandHuman ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        beta_0 : float, optional
            Value of beta_0, defaults to 2.3
        beta_1 : float, optional
            Value of beta_1, defaults to -2.4
        ca50 : float, optional
            Value of ca50, defaults to 0.805
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.08597401
        Cao : float, optional
            Value of Cao, defaults to 1.8
        cell_type : float, optional
            Value of cell_type
        Cli : float, optional
            Value of Cli, defaults to 15.0
        Clo : float, optional
            Value of Clo, defaults to 150.0
        dr : float, optional
            Value of dr, defaults to 0.25
        flags : float, optional
            Value of flags
        gamma_wu : float, optional
            Value of gamma_wu, defaults to 0.615
        gamma_zs : float, optional
            Value of gamma_zs, defaults to 0.0085
        Ki_init : float, optional
            Value of Ki_init, defaults to 120.0
        Ko : float, optional
            Value of Ko, defaults to 5.4
        koff : float, optional
            Value of koff, defaults to 0.1
        ktm_unblock : float, optional
            Value of ktm_unblock, defaults to 1.0
        lengthDep : float, optional
            Value of lengthDep, defaults to 0.0
        markov_iks : float, optional
            Value of markov_iks, defaults to 0.0
        Mgi : float, optional
            Value of Mgi, defaults to 1.0
        mu_uw : float, optional
            Value of mu_uw, defaults to 3.0
        Nai_init : float, optional
            Value of Nai_init, defaults to 9.06
        Nao : float, optional
            Value of Nao, defaults to 140.0
        nperm : float, optional
            Value of nperm, defaults to 5.0
        nu_ws : float, optional
            Value of nu_ws, defaults to 7.0
        perm50 : float, optional
            Value of perm50, defaults to 0.35
        phi : float, optional
            Value of phi, defaults to 2.23
        TOT_A : float, optional
            Value of TOT_A, defaults to 25.0
        Tref : float, optional
            Value of Tref, defaults to 120.0
        TRPN_n : float, optional
            Value of TRPN_n, defaults to 2.0
        wfrac : float, optional
            Value of wfrac, defaults to 0.5
    """
    SHORTNAME = 'gpb_landhuman'
    MODELID = 'GPB_LandHuman'
    PARAMETERS = {'beta_0': 2.3,  'beta_1': -2.4,  'ca50': 0.805,  
                  'Ca_i_init': 0.08597401,  'Cao': 1.8,  'cell_type': None,  
                  'Cli': 15.0,  'Clo': 150.0,  'dr': 0.25,  'flags': None,  
                  'gamma_wu': 0.615,  'gamma_zs': 0.0085,  'Ki_init': 120.0,  
                  'Ko': 5.4,  'koff': 0.1,  'ktm_unblock': 1.0,  
                  'lengthDep': 0.0,  'markov_iks': 0.0,  'Mgi': 1.0,  
                  'mu_uw': 3.0,  'Nai_init': 9.06,  'Nao': 140.0,  
                  'nperm': 5.0,  'nu_ws': 7.0,  'perm50': 0.35,  'phi': 2.23,  
                  'TOT_A': 25.0,  'Tref': 120.0,  'TRPN_n': 2.0,  'wfrac': 0.5}


class GW_CANIonicModel(_AbstractIonicModel):
    """
    Describes the GW_CAN ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Cao : float, optional
            Value of Cao, defaults to 2.0
        CHF : float, optional
            Value of CHF, defaults to 0.0
        flags : float, optional
            Value of flags
        G_K1 : float, optional
            Value of G_K1
        G_Ks : float, optional
            Value of G_Ks
        G_Kv43 : float, optional
            Value of G_Kv43
        gNa : float, optional
            Value of gNa, defaults to 12.8
        kNaCa : float, optional
            Value of kNaCa
        Ko : float, optional
            Value of Ko, defaults to 4.0
        ksr : float, optional
            Value of ksr
        Nao : float, optional
            Value of Nao, defaults to 138.0
        PKv14 : float, optional
            Value of PKv14, defaults to 1.708681e-07
        type : float, optional
            Value of type
    """
    SHORTNAME = 'gw_can'
    MODELID = 'GW_CAN'
    PARAMETERS = {'Cao': 2.0,  'CHF': 0.0,  'flags': None,  'G_K1': None,  
                  'G_Ks': None,  'G_Kv43': None,  'gNa': 12.8,  'kNaCa': None,  
                  'Ko': 4.0,  'ksr': None,  'Nao': 138.0,  
                  'PKv14': 1.708681e-07,  'type': None}


class hAMrIonicModel(_AbstractIonicModel):
    """
    Describes the hAMr ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        ACh : float, optional
            Value of ACh, defaults to 1e-24
        blf_i_di : float, optional
            Value of blf_i_di, defaults to 1.0
        blf_i_NaCa : float, optional
            Value of blf_i_NaCa, defaults to 1.0
        blf_i_rel : float, optional
            Value of blf_i_rel, defaults to 1.0
        blf_i_tr : float, optional
            Value of blf_i_tr, defaults to 1.0
        Cm : float, optional
            Value of Cm, defaults to 50.0
        g_B_Ca : float, optional
            Value of g_B_Ca, defaults to 0.078681
        g_B_Na : float, optional
            Value of g_B_Na, defaults to 0.060599
        g_Ca_L : float, optional
            Value of g_Ca_L, defaults to 6.75
        g_K1 : float, optional
            Value of g_K1, defaults to 3.1
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.5
        g_Ks : float, optional
            Value of g_Ks, defaults to 1.0
        g_kur : float, optional
            Value of g_kur, defaults to 2.25
        g_t : float, optional
            Value of g_t, defaults to 8.25
        i_CaP_max : float, optional
            Value of i_CaP_max, defaults to 4.0
        i_NaK_max : float, optional
            Value of i_NaK_max, defaults to 68.55
        I_up_max : float, optional
            Value of I_up_max, defaults to 2800.0
        P_ACh : float, optional
            Value of P_ACh, defaults to 1.0
        P_Na : float, optional
            Value of P_Na, defaults to 0.0018
        phi_Na_en : float, optional
            Value of phi_Na_en, defaults to -1.68
    """
    SHORTNAME = 'hamr'
    MODELID = 'hAMr'
    PARAMETERS = {'ACh': 1e-24,  'blf_i_di': 1.0,  'blf_i_NaCa': 1.0,  
                  'blf_i_rel': 1.0,  'blf_i_tr': 1.0,  'Cm': 50.0,  
                  'g_B_Ca': 0.078681,  'g_B_Na': 0.060599,  'g_Ca_L': 6.75,  
                  'g_K1': 3.1,  'g_Kr': 0.5,  'g_Ks': 1.0,  'g_kur': 2.25,  
                  'g_t': 8.25,  'i_CaP_max': 4.0,  'i_NaK_max': 68.55,  
                  'I_up_max': 2800.0,  'P_ACh': 1.0,  'P_Na': 0.0018,  
                  'phi_Na_en': -1.68}


class HHIonicModel(_AbstractIonicModel):
    """
    Describes the HH ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        g_K : float, optional
            Value of g_K, defaults to 36.0
        g_L : float, optional
            Value of g_L, defaults to 0.3
        g_Na : float, optional
            Value of g_Na, defaults to 120.0
    """
    SHORTNAME = 'hh'
    MODELID = 'HH'
    PARAMETERS = {'g_K': 36.0,  'g_L': 0.3,  'g_Na': 120.0}


class IION_SRCIonicModel(_AbstractIonicModel):
    """
    Describes the IION_SRC ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        AP_ID : float, optional
            Value of AP_ID, defaults to 0.0
    """
    SHORTNAME = 'iion_src'
    MODELID = 'IION_SRC'
    PARAMETERS = {'AP_ID': 0.0}


class INADA_AVNIonicModel(_AbstractIonicModel):
    """
    Describes the INADA_AVN ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        cell_type : float, optional
            Value of cell_type
        Cm_cell : float, optional
            Value of Cm_cell
        E_b : float, optional
            Value of E_b
        E_CaL : float, optional
            Value of E_CaL
        flags : float, optional
            Value of flags
        g_b : float, optional
            Value of g_b
        g_CaL : float, optional
            Value of g_CaL
        g_f : float, optional
            Value of g_f
        g_K1 : float, optional
            Value of g_K1
        g_Kr : float, optional
            Value of g_Kr
        g_Na : float, optional
            Value of g_Na
        g_st : float, optional
            Value of g_st
        g_to : float, optional
            Value of g_to
        k_NaCa : float, optional
            Value of k_NaCa
        p_max : float, optional
            Value of p_max
        p_rel : float, optional
            Value of p_rel
    """
    SHORTNAME = 'inada_avn'
    MODELID = 'INADA_AVN'
    PARAMETERS = {'cell_type': None,  'Cm_cell': None,  'E_b': None,  
                  'E_CaL': None,  'flags': None,  'g_b': None,  'g_CaL': None,  
                  'g_f': None,  'g_K1': None,  'g_Kr': None,  'g_Na': None,  
                  'g_st': None,  'g_to': None,  'k_NaCa': None,  'p_max': None,  
                  'p_rel': None}


class IribeIonicModel(_AbstractIonicModel):
    """
    Describes the Iribe ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        K_i_init : float, optional
            Value of K_i_init, defaults to 138.22
    """
    SHORTNAME = 'iribe'
    MODELID = 'Iribe'
    PARAMETERS = {'K_i_init': 138.22}


class KurataIonicModel(_AbstractIonicModel):
    """
    Describes the Kurata ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        ACh : float, optional
            Value of ACh, defaults to 0.0
        B_max : float, optional
            Value of B_max, defaults to 0.56
        Cm : float, optional
            Value of Cm, defaults to 32.0
        G_CaL : float, optional
            Value of G_CaL, defaults to 0.58
        G_h : float, optional
            Value of G_h, defaults to 0.357
        G_KACh_max : float, optional
            Value of G_KACh_max
        G_Kr : float, optional
            Value of G_Kr
        G_Ks : float, optional
            Value of G_Ks, defaults to 0.025
        G_Na : float, optional
            Value of G_Na, defaults to 0.0
        G_st : float, optional
            Value of G_st, defaults to 0.015
        G_sus : float, optional
            Value of G_sus, defaults to 0.02
        G_to : float, optional
            Value of G_to, defaults to 0.18
    """
    SHORTNAME = 'kurata'
    MODELID = 'Kurata'
    PARAMETERS = {'ACh': 0.0,  'B_max': 0.56,  'Cm': 32.0,  'G_CaL': 0.58,  
                  'G_h': 0.357,  'G_KACh_max': None,  'G_Kr': None,  
                  'G_Ks': 0.025,  'G_Na': 0.0,  'G_st': 0.015,  'G_sus': 0.02,  
                  'G_to': 0.18}


class LMCGIonicModel(_AbstractIonicModel):
    """
    Describes the LMCG ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Cao : float, optional
            Value of Cao, defaults to 2.5
        cell_type : float, optional
            Value of cell_type
        ECaL : float, optional
            Value of ECaL, defaults to 50.0
        flags : float, optional
            Value of flags
        gCaB : float, optional
            Value of gCaB, defaults to 0.03
        gCaL : float, optional
            Value of gCaL, defaults to 4.0
        gCaP : float, optional
            Value of gCaP, defaults to 9.509
        gCaT : float, optional
            Value of gCaT, defaults to 6.0
        gClB : float, optional
            Value of gClB, defaults to 0.0
        gK1 : float, optional
            Value of gK1, defaults to 10.0
        gKr : float, optional
            Value of gKr, defaults to 7.0
        gKs : float, optional
            Value of gKs, defaults to 5.0
        gNaB : float, optional
            Value of gNaB, defaults to 0.03
        gNaCa : float, optional
            Value of gNaCa, defaults to 0.02
        gsus : float, optional
            Value of gsus, defaults to 2.4
        gt : float, optional
            Value of gt, defaults to 35.0
        iNaKmax : float, optional
            Value of iNaKmax, defaults to 64.41
        Ki : float, optional
            Value of Ki, defaults to 139.88
        Ko : float, optional
            Value of Ko, defaults to 5.0
        Nai : float, optional
            Value of Nai, defaults to 8.438
        Nao : float, optional
            Value of Nao, defaults to 140.0
        pNa : float, optional
            Value of pNa, defaults to 0.0014
    """
    SHORTNAME = 'lmcg'
    MODELID = 'LMCG'
    PARAMETERS = {'Cao': 2.5,  'cell_type': None,  'ECaL': 50.0,  
                  'flags': None,  'gCaB': 0.03,  'gCaL': 4.0,  'gCaP': 9.509,  
                  'gCaT': 6.0,  'gClB': 0.0,  'gK1': 10.0,  'gKr': 7.0,  
                  'gKs': 5.0,  'gNaB': 0.03,  'gNaCa': 0.02,  'gsus': 2.4,  
                  'gt': 35.0,  'iNaKmax': 64.41,  'Ki': 139.88,  'Ko': 5.0,  
                  'Nai': 8.438,  'Nao': 140.0,  'pNa': 0.0014}


class LR1IonicModel(_AbstractIonicModel):
    """
    Describes the LR1 ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.0002
        Gb : float, optional
            Value of Gb, defaults to 0.03921
        GK1bar : float, optional
            Value of GK1bar, defaults to 0.604
        GKbar : float, optional
            Value of GKbar, defaults to 0.282
        GKp : float, optional
            Value of GKp, defaults to 0.0183
        GNa : float, optional
            Value of GNa, defaults to 23.0
        Gsi : float, optional
            Value of Gsi, defaults to 0.09
        Ki : float, optional
            Value of Ki, defaults to 145.0
        Ko : float, optional
            Value of Ko, defaults to 5.4
        Nai : float, optional
            Value of Nai, defaults to 18.0
        Nao : float, optional
            Value of Nao, defaults to 140.0
        tau_d_factor : float, optional
            Value of tau_d_factor, defaults to 1.0
        tau_f_factor : float, optional
            Value of tau_f_factor, defaults to 1.0
    """
    SHORTNAME = 'lr1'
    MODELID = 'LR1'
    PARAMETERS = {'Ca_i_init': 0.0002,  'Gb': 0.03921,  'GK1bar': 0.604,  
                  'GKbar': 0.282,  'GKp': 0.0183,  'GNa': 23.0,  'Gsi': 0.09,  
                  'Ki': 145.0,  'Ko': 5.4,  'Nai': 18.0,  'Nao': 140.0,  
                  'tau_d_factor': 1.0,  'tau_f_factor': 1.0}


class LRDIIIonicModel(_AbstractIonicModel):
    """
    Describes the LRDII ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        *capulse_AP : float, optional
            Value of *capulse_AP
        *capulse_CaO : float, optional
            Value of *capulse_CaO
        *flags : float, optional
            Value of *flags
        caincyc : float, optional
            Value of caincyc
        cell_type : float, optional
            Value of cell_type, defaults to 1.0
        dvdt_thresh : float, optional
            Value of dvdt_thresh, defaults to 20.0
        extern_Cab : float, optional
            Value of extern_Cab, defaults to 0.0
        gCaL : float, optional
            Value of gCaL, defaults to 1.0
        gKr : float, optional
            Value of gKr, defaults to 0.02614
        gKs : float, optional
            Value of gKs, defaults to 1.0
        grel_AP : float, optional
            Value of grel_AP, defaults to 38.0
        grel_CaO : float, optional
            Value of grel_CaO, defaults to 3.0
        pulsecyc : float, optional
            Value of pulsecyc
        ratio[0] : float, optional
            Value of ratio[0], defaults to 1.0
        ratio[1] : float, optional
            Value of ratio[1], defaults to 1.0
        ratio[2] : float, optional
            Value of ratio[2], defaults to 1.0
        ratio[3] : float, optional
            Value of ratio[3]
        tauSR_AP : float, optional
            Value of tauSR_AP, defaults to 3.5
        tauSR_CaO : float, optional
            Value of tauSR_CaO, defaults to 6.0
    """
    SHORTNAME = 'lrdii'
    MODELID = 'LRDII'
    PARAMETERS = {'*capulse_AP': None,  '*capulse_CaO': None,  '*flags': None,  
                  'caincyc': None,  'cell_type': 1.0,  'dvdt_thresh': 20.0,  
                  'extern_Cab': 0.0,  'gCaL': 1.0,  'gKr': 0.02614,  
                  'gKs': 1.0,  'grel_AP': 38.0,  'grel_CaO': 3.0,  
                  'pulsecyc': None,  'ratio[0]': 1.0,  'ratio[1]': 1.0,  
                  'ratio[2]': 1.0,  'ratio[3]': None,  'tauSR_AP': 3.5,  
                  'tauSR_CaO': 6.0}


class LRDII_FIonicModel(_AbstractIonicModel):
    """
    Describes the LRDII_F ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        *capulse_cicr : float, optional
            Value of *capulse_cicr
        *capulse_jsrol : float, optional
            Value of *capulse_jsrol
        cicr_pulsecyc : float, optional
            Value of cicr_pulsecyc
        ext_conc_dt : float, optional
            Value of ext_conc_dt, defaults to 0.0
        extern_Cab : float, optional
            Value of extern_Cab
        g_NaB : float, optional
            Value of g_NaB, defaults to 0.004
        g_NaK : float, optional
            Value of g_NaK, defaults to 1.0
        gamma : float, optional
            Value of gamma, defaults to 0.8
        jsrol_pulsecyc : float, optional
            Value of jsrol_pulsecyc
        scl_tau_f : float, optional
            Value of scl_tau_f, defaults to 1.0
    """
    SHORTNAME = 'lrdii_f'
    MODELID = 'LRDII_F'
    PARAMETERS = {'*capulse_cicr': None,  '*capulse_jsrol': None,  
                  'cicr_pulsecyc': None,  'ext_conc_dt': 0.0,  
                  'extern_Cab': None,  'g_NaB': 0.004,  'g_NaK': 1.0,  
                  'gamma': 0.8,  'jsrol_pulsecyc': None,  'scl_tau_f': 1.0}


class MacCannell_FbIonicModel(_AbstractIonicModel):
    """
    Describes the MacCannell_Fb ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        g_b_Na : float, optional
            Value of g_b_Na, defaults to 0.0095
        g_K1 : float, optional
            Value of g_K1, defaults to 0.4822
        g_Kv : float, optional
            Value of g_Kv, defaults to 0.25
        i_bar_NaK : float, optional
            Value of i_bar_NaK, defaults to 1.4
        K_i_init : float, optional
            Value of K_i_init, defaults to 140.045865
        K_o : float, optional
            Value of K_o, defaults to 5.4
        Na_i_init : float, optional
            Value of Na_i_init, defaults to 9.954213
        Na_o : float, optional
            Value of Na_o, defaults to 130.0
    """
    SHORTNAME = 'maccannell_fb'
    MODELID = 'MacCannell_Fb'
    PARAMETERS = {'g_b_Na': 0.0095,  'g_K1': 0.4822,  'g_Kv': 0.25,  
                  'i_bar_NaK': 1.4,  'K_i_init': 140.045865,  'K_o': 5.4,  
                  'Na_i_init': 9.954213,  'Na_o': 130.0}


class MitchellSchaefferIonicModel(_AbstractIonicModel):
    """
    Describes the MitchellSchaeffer ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        tau_close : float, optional
            Value of tau_close, defaults to 150.0
        tau_in : float, optional
            Value of tau_in, defaults to 0.3
        tau_open : float, optional
            Value of tau_open, defaults to 120.0
        tau_out : float, optional
            Value of tau_out, defaults to 5.0
        V_gate : float, optional
            Value of V_gate, defaults to 0.13
        V_max : float, optional
            Value of V_max, defaults to 1.0
        V_min : float, optional
            Value of V_min, defaults to 0.0
    """
    SHORTNAME = 'mitchellschaeffer'
    MODELID = 'MitchellSchaeffer'
    PARAMETERS = {'tau_close': 150.0,  'tau_in': 0.3,  'tau_open': 120.0,  
                  'tau_out': 5.0,  'V_gate': 0.13,  'V_max': 1.0,  
                  'V_min': 0.0}


class mMSIonicModel(_AbstractIonicModel):
    """
    Describes the mMS ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        a_crit : float, optional
            Value of a_crit, defaults to 0.13
        tau_close : float, optional
            Value of tau_close, defaults to 150.0
        tau_in : float, optional
            Value of tau_in, defaults to 0.3
        tau_open : float, optional
            Value of tau_open, defaults to 120.0
        tau_out : float, optional
            Value of tau_out, defaults to 5.0
        V_gate : float, optional
            Value of V_gate, defaults to 0.13
        V_max : float, optional
            Value of V_max, defaults to 1.0
        V_min : float, optional
            Value of V_min, defaults to 0.0
    """
    SHORTNAME = 'mms'
    MODELID = 'mMS'
    PARAMETERS = {'a_crit': 0.13,  'tau_close': 150.0,  'tau_in': 0.3,  
                  'tau_open': 120.0,  'tau_out': 5.0,  'V_gate': 0.13,  
                  'V_max': 1.0,  'V_min': 0.0}


class MBRDRIonicModel(_AbstractIonicModel):
    """
    Describes the MBRDR ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        APDshorten : float, optional
            Value of APDshorten, defaults to 1.0
        GNa : float, optional
            Value of GNa, defaults to 15.0
        Gsi : float, optional
            Value of Gsi, defaults to 0.09
    """
    SHORTNAME = 'mbrdr'
    MODELID = 'MBRDR'
    PARAMETERS = {'APDshorten': 1.0,  'GNa': 15.0,  'Gsi': 0.09}


class MurineMouseIonicModel(_AbstractIonicModel):
    """
    Describes the MurineMouse ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        a : float, optional
            Value of a, defaults to 0.0625
        Acap : float, optional
            Value of Acap, defaults to 0.00013866
        aKss_init : float, optional
            Value of aKss_init, defaults to 0.287585636847048
        ato_f_init : float, optional
            Value of ato_f_init, defaults to 0.0142335908879204
        ato_s_init : float, optional
            Value of ato_s_init, defaults to 0.0443263407760382
        aur_init : float, optional
            Value of aur_init, defaults to 0.00346258606821817
        b : float, optional
            Value of b, defaults to 0.4
        Beta_Na4 : float, optional
            Value of Beta_Na4
        Beta_Na5 : float, optional
            Value of Beta_Na5
        Bmax : float, optional
            Value of Bmax, defaults to 109.0
        C_K1_init : float, optional
            Value of C_K1_init, defaults to 0.0011733433957123
        C_K2_init : float, optional
            Value of C_K2_init, defaults to 0.00105586824723736
        C_Na1_init : float, optional
            Value of C_Na1_init, defaults to 0.000420472760277688
        C_Na2_init : float, optional
            Value of C_Na2_init, defaults to 0.0240114508843199
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.0949915068139801
        CaJSR_init : float, optional
            Value of CaJSR_init, defaults to 171.167969039613
        CaMKt_init : float, optional
            Value of CaMKt_init, defaults to 0.729027738385079
        CaNSR_init : float, optional
            Value of CaNSR_init, defaults to 404.825013216286
        Cao : float, optional
            Value of Cao, defaults to 1400.0
        Cass_init : float, optional
            Value of Cass_init, defaults to 0.0954184301907784
        Cm : float, optional
            Value of Cm, defaults to 1.0
        const5 : float, optional
            Value of const5, defaults to 8.2
        CSQN_tot : float, optional
            Value of CSQN_tot, defaults to 50000.0
        delta_V_L : float, optional
            Value of delta_V_L, defaults to 6.4489
        E_Cl : float, optional
            Value of E_Cl, defaults to -40.0
        eta : float, optional
            Value of eta, defaults to 0.35
        F : float, optional
            Value of F, defaults to 96.5
        g_Cab : float, optional
            Value of g_Cab, defaults to 0.0007
        g_ClCa : float, optional
            Value of g_ClCa, defaults to 10.0
        g_K1 : float, optional
            Value of g_K1, defaults to 0.35
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0165
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.00575
        g_Kss : float, optional
            Value of g_Kss, defaults to 0.0595
        g_Kto_f : float, optional
            Value of g_Kto_f, defaults to 0.5347
        g_Kto_s : float, optional
            Value of g_Kto_s, defaults to 0.0
        g_Kur : float, optional
            Value of g_Kur, defaults to 0.25
        g_Na : float, optional
            Value of g_Na, defaults to 16.0
        g_Nab : float, optional
            Value of g_Nab, defaults to 0.0026
        I1_Na_init : float, optional
            Value of I1_Na_init, defaults to 0.000517471697712382
        I2_Na_init : float, optional
            Value of I2_Na_init, defaults to 2.45406116958509e-05
        i_CaL_max : float, optional
            Value of i_CaL_max, defaults to 7.0
        I_init : float, optional
            Value of I_init, defaults to 0.427651445872853
        I_K_init : float, optional
            Value of I_K_init, defaults to 0.00140618453684944
        i_NaK_max : float, optional
            Value of i_NaK_max, defaults to 2.486
        i_pCa_max : float, optional
            Value of i_pCa_max, defaults to 0.0955
        IC_Na2_init : float, optional
            Value of IC_Na2_init, defaults to 0.0174528857380179
        IC_Na3_init : float, optional
            Value of IC_Na3_init, defaults to 0.402980726914811
        IF_Na_init : float, optional
            Value of IF_Na_init, defaults to 0.000306123648969581
        ito_f_init : float, optional
            Value of ito_f_init, defaults to 0.996989882138174
        ito_s_init : float, optional
            Value of ito_s_init, defaults to 0.887568880831388
        iur_init : float, optional
            Value of iur_init, defaults to 0.955684946168062
        K_L : float, optional
            Value of K_L, defaults to 0.3
        K_mAllo : float, optional
            Value of K_mAllo, defaults to 0.0
        K_mCai : float, optional
            Value of K_mCai, defaults to 3.6
        K_mCao : float, optional
            Value of K_mCao, defaults to 1400.0
        k_minus_a : float, optional
            Value of k_minus_a, defaults to 0.07125
        k_minus_b : float, optional
            Value of k_minus_b, defaults to 0.965
        k_minus_c : float, optional
            Value of k_minus_c, defaults to 0.0008
        K_mNai : float, optional
            Value of K_mNai, defaults to 12000.0
        K_mNao : float, optional
            Value of K_mNao, defaults to 88000.0
        k_plus_a : float, optional
            Value of k_plus_a, defaults to 0.006075
        k_plus_b : float, optional
            Value of k_plus_b, defaults to 0.00405
        k_plus_c : float, optional
            Value of k_plus_c, defaults to 0.009
        k_sat : float, optional
            Value of k_sat, defaults to 0.27
        kb : float, optional
            Value of kb, defaults to 0.036778
        Kd : float, optional
            Value of Kd, defaults to 0.6
        kf : float, optional
            Value of kf, defaults to 0.023761
        Ki_init : float, optional
            Value of Ki_init, defaults to 115599.50642567
        Km_Cl : float, optional
            Value of Km_Cl, defaults to 10.0
        Km_CSQN : float, optional
            Value of Km_CSQN, defaults to 630.0
        Km_Ko : float, optional
            Value of Km_Ko, defaults to 1500.0
        Km_Nai : float, optional
            Value of Km_Nai, defaults to 16600.0
        Km_pCa : float, optional
            Value of Km_pCa, defaults to 0.2885
        Km_up : float, optional
            Value of Km_up, defaults to 0.412
        Ko : float, optional
            Value of Ko, defaults to 5400.0
        m : float, optional
            Value of m, defaults to 3.0
        n : float, optional
            Value of n, defaults to 4.0
        Nai_init : float, optional
            Value of Nai_init, defaults to 12364.7482121793
        Nao : float, optional
            Value of Nao, defaults to 134000.0
        nKs_init : float, optional
            Value of nKs_init, defaults to 0.00336735013094628
        O_init : float, optional
            Value of O_init, defaults to 1.23713515513533e-06
        O_K_init : float, optional
            Value of O_K_init, defaults to 0.0131742086125972
        O_Na_init : float, optional
            Value of O_Na_init, defaults to 1.46826771436314e-06
        off_rate : float, optional
            Value of off_rate, defaults to 0.0002
        on_rate : float, optional
            Value of on_rate, defaults to 0.05
        P_C2_init : float, optional
            Value of P_C2_init, defaults to 0.565182571165673
        P_CaL : float, optional
            Value of P_CaL, defaults to 2.5
        P_O1_init : float, optional
            Value of P_O1_init, defaults to 0.00571393383393735
        P_O2_init : float, optional
            Value of P_O2_init, defaults to 2.09864618235341e-08
        P_ryr_const1 : float, optional
            Value of P_ryr_const1, defaults to -0.01
        P_ryr_const2 : float, optional
            Value of P_ryr_const2, defaults to -2.0
        P_RyR_init : float, optional
            Value of P_RyR_init, defaults to 0.000280539508743811
        phi_L : float, optional
            Value of phi_L, defaults to 1.798
        R : float, optional
            Value of R, defaults to 8.314
        T : float, optional
            Value of T, defaults to 308.0
        t_L : float, optional
            Value of t_L, defaults to 1.5
        tau_i_const : float, optional
            Value of tau_i_const, defaults to 643.0
        tau_L : float, optional
            Value of tau_L, defaults to 1150.0
        tau_tr : float, optional
            Value of tau_tr, defaults to 20.0
        tau_xfer : float, optional
            Value of tau_xfer, defaults to 8.0
        v1 : float, optional
            Value of v1, defaults to 4.5
        v1_caff : float, optional
            Value of v1_caff, defaults to 10.0
        v2 : float, optional
            Value of v2, defaults to 3e-05
        v2_caff : float, optional
            Value of v2_caff, defaults to 0.1
        V_init : float, optional
            Value of V_init, defaults to -78.9452115785979
        V_L : float, optional
            Value of V_L, defaults to 0.0
        V_max_NCX : float, optional
            Value of V_max_NCX, defaults to 3.939
        VJSR : float, optional
            Value of VJSR, defaults to 7.7e-08
        vmup_init : float, optional
            Value of vmup_init, defaults to 0.5059
        Vmyo : float, optional
            Value of Vmyo, defaults to 2.2e-05
        VNSR : float, optional
            Value of VNSR, defaults to 2.31e-07
        Vss : float, optional
            Value of Vss, defaults to 2.2e-08
        y_gate_init : float, optional
            Value of y_gate_init, defaults to 0.845044436980163
        y_gate_tau_const1 : float, optional
            Value of y_gate_tau_const1, defaults to 8.0
        y_gate_tau_const2 : float, optional
            Value of y_gate_tau_const2, defaults to 315.0
    """
    SHORTNAME = 'murinemouse'
    MODELID = 'MurineMouse'
    PARAMETERS = {'a': 0.0625,  'Acap': 0.00013866,  
                  'aKss_init': 0.287585636847048,  
                  'ato_f_init': 0.0142335908879204,  
                  'ato_s_init': 0.0443263407760382,  
                  'aur_init': 0.00346258606821817,  'b': 0.4,  
                  'Beta_Na4': None,  'Beta_Na5': None,  'Bmax': 109.0,  
                  'C_K1_init': 0.0011733433957123,  
                  'C_K2_init': 0.00105586824723736,  
                  'C_Na1_init': 0.000420472760277688,  
                  'C_Na2_init': 0.0240114508843199,  
                  'Ca_i_init': 0.0949915068139801,  
                  'CaJSR_init': 171.167969039613,  
                  'CaMKt_init': 0.729027738385079,  
                  'CaNSR_init': 404.825013216286,  'Cao': 1400.0,  
                  'Cass_init': 0.0954184301907784,  'Cm': 1.0,  'const5': 8.2,  
                  'CSQN_tot': 50000.0,  'delta_V_L': 6.4489,  'E_Cl': -40.0,  
                  'eta': 0.35,  'F': 96.5,  'g_Cab': 0.0007,  'g_ClCa': 10.0,  
                  'g_K1': 0.35,  'g_Kr': 0.0165,  'g_Ks': 0.00575,  
                  'g_Kss': 0.0595,  'g_Kto_f': 0.5347,  'g_Kto_s': 0.0,  
                  'g_Kur': 0.25,  'g_Na': 16.0,  'g_Nab': 0.0026,  
                  'I1_Na_init': 0.000517471697712382,  
                  'I2_Na_init': 2.45406116958509e-05,  'i_CaL_max': 7.0,  
                  'I_init': 0.427651445872853,  
                  'I_K_init': 0.00140618453684944,  'i_NaK_max': 2.486,  
                  'i_pCa_max': 0.0955,  'IC_Na2_init': 0.0174528857380179,  
                  'IC_Na3_init': 0.402980726914811,  
                  'IF_Na_init': 0.000306123648969581,  
                  'ito_f_init': 0.996989882138174,  
                  'ito_s_init': 0.887568880831388,  
                  'iur_init': 0.955684946168062,  'K_L': 0.3,  'K_mAllo': 0.0,  
                  'K_mCai': 3.6,  'K_mCao': 1400.0,  'k_minus_a': 0.07125,  
                  'k_minus_b': 0.965,  'k_minus_c': 0.0008,  'K_mNai': 12000.0,  
                  'K_mNao': 88000.0,  'k_plus_a': 0.006075,  
                  'k_plus_b': 0.00405,  'k_plus_c': 0.009,  'k_sat': 0.27,  
                  'kb': 0.036778,  'Kd': 0.6,  'kf': 0.023761,  
                  'Ki_init': 115599.50642567,  'Km_Cl': 10.0,  
                  'Km_CSQN': 630.0,  'Km_Ko': 1500.0,  'Km_Nai': 16600.0,  
                  'Km_pCa': 0.2885,  'Km_up': 0.412,  'Ko': 5400.0,  'm': 3.0,  
                  'n': 4.0,  'Nai_init': 12364.7482121793,  'Nao': 134000.0,  
                  'nKs_init': 0.00336735013094628,  
                  'O_init': 1.23713515513533e-06,  
                  'O_K_init': 0.0131742086125972,  
                  'O_Na_init': 1.46826771436314e-06,  'off_rate': 0.0002,  
                  'on_rate': 0.05,  'P_C2_init': 0.565182571165673,  
                  'P_CaL': 2.5,  'P_O1_init': 0.00571393383393735,  
                  'P_O2_init': 2.09864618235341e-08,  'P_ryr_const1': -0.01,  
                  'P_ryr_const2': -2.0,  'P_RyR_init': 0.000280539508743811,  
                  'phi_L': 1.798,  'R': 8.314,  'T': 308.0,  't_L': 1.5,  
                  'tau_i_const': 643.0,  'tau_L': 1150.0,  'tau_tr': 20.0,  
                  'tau_xfer': 8.0,  'v1': 4.5,  'v1_caff': 10.0,  'v2': 3e-05,  
                  'v2_caff': 0.1,  'V_init': -78.9452115785979,  'V_L': 0.0,  
                  'V_max_NCX': 3.939,  'VJSR': 7.7e-08,  'vmup_init': 0.5059,  
                  'Vmyo': 2.2e-05,  'VNSR': 2.31e-07,  'Vss': 2.2e-08,  
                  'y_gate_init': 0.845044436980163,  'y_gate_tau_const1': 8.0,  
                  'y_gate_tau_const2': 315.0}


class NYGRENIonicModel(_AbstractIonicModel):
    """
    Describes the NYGREN ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        extern_Cab : float, optional
            Value of extern_Cab
        G_Ca : float, optional
            Value of G_Ca, defaults to 1.0
        G_Kr : float, optional
            Value of G_Kr, defaults to 1.0
        G_to : float, optional
            Value of G_to, defaults to 1.0
    """
    SHORTNAME = 'nygren'
    MODELID = 'NYGREN'
    PARAMETERS = {'extern_Cab': None,  'G_Ca': 1.0,  'G_Kr': 1.0,  'G_to': 1.0}


class ORdIonicModel(_AbstractIonicModel):
    """
    Describes the ORd ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        blf_ICab : float, optional
            Value of blf_ICab, defaults to 1.0
        blf_ICaK : float, optional
            Value of blf_ICaK, defaults to 1.0
        blf_ICaL : float, optional
            Value of blf_ICaL, defaults to 1.0
        blf_ICaNa : float, optional
            Value of blf_ICaNa, defaults to 1.0
        blf_INab : float, optional
            Value of blf_INab, defaults to 1.0
        blf_INaCa_i : float, optional
            Value of blf_INaCa_i, defaults to 1.0
        blf_INaCa_ss : float, optional
            Value of blf_INaCa_ss, defaults to 1.0
        blf_INaK : float, optional
            Value of blf_INaK, defaults to 1.0
        cao : float, optional
            Value of cao, defaults to 1.8
        celltype : float, optional
            Value of celltype
        flags : float, optional
            Value of flags
        GK1 : float, optional
            Value of GK1
        GKb : float, optional
            Value of GKb
        GKr : float, optional
            Value of GKr
        GKs : float, optional
            Value of GKs
        GNa : float, optional
            Value of GNa
        GNaL : float, optional
            Value of GNaL
        Gncx : float, optional
            Value of Gncx
        GpCa : float, optional
            Value of GpCa, defaults to 0.0005
        Gto : float, optional
            Value of Gto
        INa_Type : float, optional
            Value of INa_Type
        jrel_stiff_const : float, optional
            Value of jrel_stiff_const, defaults to 0.005
        ko : float, optional
            Value of ko, defaults to 5.4
        nao : float, optional
            Value of nao, defaults to 140.0
        scale_tau_m : float, optional
            Value of scale_tau_m, defaults to 0.3
        shift_m_inf : float, optional
            Value of shift_m_inf, defaults to -8.0
    """
    SHORTNAME = 'ord'
    MODELID = 'ORd'
    PARAMETERS = {'blf_ICab': 1.0,  'blf_ICaK': 1.0,  'blf_ICaL': 1.0,  
                  'blf_ICaNa': 1.0,  'blf_INab': 1.0,  'blf_INaCa_i': 1.0,  
                  'blf_INaCa_ss': 1.0,  'blf_INaK': 1.0,  'cao': 1.8,  
                  'celltype': None,  'flags': None,  'GK1': None,  'GKb': None,  
                  'GKr': None,  'GKs': None,  'GNa': None,  'GNaL': None,  
                  'Gncx': None,  'GpCa': 0.0005,  'Gto': None,  
                  'INa_Type': None,  'jrel_stiff_const': 0.005,  'ko': 5.4,  
                  'nao': 140.0,  'scale_tau_m': 0.3,  'shift_m_inf': -8.0}


class PanditIonicModel(_AbstractIonicModel):
    """
    Describes the Pandit ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        *flags : float, optional
            Value of *flags
        a : float, optional
            Value of a, defaults to 0.886
        b : float, optional
            Value of b
        Beta : float, optional
            Value of Beta
        C_m : float, optional
            Value of C_m
        cell_type : float, optional
            Value of cell_type, defaults to 0.0
        concFact : float, optional
            Value of concFact, defaults to 1.0
        EGTA : float, optional
            Value of EGTA, defaults to 0.0
        G_CaB : float, optional
            Value of G_CaB, defaults to 3.24e-05
        G_CaL : float, optional
            Value of G_CaL, defaults to 0.031
        G_Na : float, optional
            Value of G_Na, defaults to 0.8
        G_NaB : float, optional
            Value of G_NaB, defaults to 8.015e-05
        G_SS : float, optional
            Value of G_SS, defaults to 0.007
        G_to : float, optional
            Value of G_to, defaults to 0.035
        I_NaK_ss : float, optional
            Value of I_NaK_ss, defaults to 0.08
        K1fact : float, optional
            Value of K1fact, defaults to 1.0
        K_O : float, optional
            Value of K_O, defaults to 5.4
        K_sr : float, optional
            Value of K_sr
        Na_O : float, optional
            Value of Na_O, defaults to 140.0
        Nu_maxf : float, optional
            Value of Nu_maxf, defaults to 0.04
        srMethod : float, optional
            Value of srMethod
        VclampF : float, optional
            Value of VclampF, defaults to 0.0
        VclampV : float, optional
            Value of VclampV, defaults to 0.0
    """
    SHORTNAME = 'pandit'
    MODELID = 'Pandit'
    PARAMETERS = {'*flags': None,  'a': 0.886,  'b': None,  'Beta': None,  
                  'C_m': None,  'cell_type': 0.0,  'concFact': 1.0,  
                  'EGTA': 0.0,  'G_CaB': 3.24e-05,  'G_CaL': 0.031,  
                  'G_Na': 0.8,  'G_NaB': 8.015e-05,  'G_SS': 0.007,  
                  'G_to': 0.035,  'I_NaK_ss': 0.08,  'K1fact': 1.0,  
                  'K_O': 5.4,  'K_sr': None,  'Na_O': 140.0,  'Nu_maxf': 0.04,  
                  'srMethod': None,  'VclampF': 0.0,  'VclampV': 0.0}


class PASSIVEIonicModel(_AbstractIonicModel):
    """
    Describes the PASSIVE ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Gm : float, optional
            Value of Gm
        Rm : float, optional
            Value of Rm, defaults to 10.0
        Vrest : float, optional
            Value of Vrest, defaults to 0.0
    """
    SHORTNAME = 'passive'
    MODELID = 'PASSIVE'
    PARAMETERS = {'Gm': None,  'Rm': 10.0,  'Vrest': 0.0}


class PUGIonicModel(_AbstractIonicModel):
    """
    Describes the PUG ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        *capulse_cicr : float, optional
            Value of *capulse_cicr
        *capulse_jsrol : float, optional
            Value of *capulse_jsrol
        *flags : float, optional
            Value of *flags
        Cai_init : float, optional
            Value of Cai_init
        CaNSRbar : float, optional
            Value of CaNSRbar
        Cao : float, optional
            Value of Cao
        cell_type : float, optional
            Value of cell_type, defaults to 0.0
        cicr_pulsecyc : float, optional
            Value of cicr_pulsecyc
        Cm : float, optional
            Value of Cm
        dvdt_thresh : float, optional
            Value of dvdt_thresh, defaults to 20.0
        ECaT : float, optional
            Value of ECaT
        extern_Cab : float, optional
            Value of extern_Cab
        ff_GCaL : float, optional
            Value of ff_GCaL
        ff_GKs : float, optional
            Value of ff_GKs
        ff_INaCa : float, optional
            Value of ff_INaCa
        ff_tauXs : float, optional
            Value of ff_tauXs
        GCab : float, optional
            Value of GCab
        GCaT : float, optional
            Value of GCaT, defaults to 0.05
        GClCa : float, optional
            Value of GClCa
        GK1 : float, optional
            Value of GK1, defaults to 0.54
        GKp : float, optional
            Value of GKp, defaults to 0.008
        GKr : float, optional
            Value of GKr, defaults to 0.035
        GNa : float, optional
            Value of GNa, defaults to 8.0
        GNab : float, optional
            Value of GNab
        Gto : float, optional
            Value of Gto, defaults to 0.06
        Iupbar : float, optional
            Value of Iupbar
        jsrol_pulsecyc : float, optional
            Value of jsrol_pulsecyc
        Ki_const : float, optional
            Value of Ki_const
        Ki_init : float, optional
            Value of Ki_init
        Kmup : float, optional
            Value of Kmup
        Ko : float, optional
            Value of Ko
        Nai_const : float, optional
            Value of Nai_const
        Nai_init : float, optional
            Value of Nai_init
        ratio[0] : float, optional
            Value of ratio[0], defaults to 1.0
        ratio[1] : float, optional
            Value of ratio[1], defaults to 1.0
        ratio[2] : float, optional
            Value of ratio[2], defaults to 1.0
        ratio[3] : float, optional
            Value of ratio[3]
        T : float, optional
            Value of T
        tau_tr : float, optional
            Value of tau_tr
        validate : float, optional
            Value of validate
        Vc : float, optional
            Value of Vc
        Vm_init : float, optional
            Value of Vm_init
    """
    SHORTNAME = 'pug'
    MODELID = 'PUG'
    PARAMETERS = {'*capulse_cicr': None,  '*capulse_jsrol': None,  
                  '*flags': None,  'Cai_init': None,  'CaNSRbar': None,  
                  'Cao': None,  'cell_type': 0.0,  'cicr_pulsecyc': None,  
                  'Cm': None,  'dvdt_thresh': 20.0,  'ECaT': None,  
                  'extern_Cab': None,  'ff_GCaL': None,  'ff_GKs': None,  
                  'ff_INaCa': None,  'ff_tauXs': None,  'GCab': None,  
                  'GCaT': 0.05,  'GClCa': None,  'GK1': 0.54,  'GKp': 0.008,  
                  'GKr': 0.035,  'GNa': 8.0,  'GNab': None,  'Gto': 0.06,  
                  'Iupbar': None,  'jsrol_pulsecyc': None,  'Ki_const': None,  
                  'Ki_init': None,  'Kmup': None,  'Ko': None,  
                  'Nai_const': None,  'Nai_init': None,  'ratio[0]': 1.0,  
                  'ratio[1]': 1.0,  'ratio[2]': 1.0,  'ratio[3]': None,  
                  'T': None,  'tau_tr': None,  'validate': None,  'Vc': None,  
                  'Vm_init': None}


class Rat_neonIonicModel(_AbstractIonicModel):
    """
    Describes the Rat_neon ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        AcapCm2F : float, optional
            Value of AcapCm2F
        AcapCmFVmyo : float, optional
            Value of AcapCmFVmyo
        CMDNtot : float, optional
            Value of CMDNtot, defaults to 50.0
        CSQNtot : float, optional
            Value of CSQNtot, defaults to 24750.0
        DCa : float, optional
            Value of DCa, defaults to 7.0
        dx : float, optional
            Value of dx, defaults to 0.1
        GCaL : float, optional
            Value of GCaL
        GCaT : float, optional
            Value of GCaT, defaults to 0.2
        Gf : float, optional
            Value of Gf, defaults to 0.021
        GK1 : float, optional
            Value of GK1, defaults to 0.0515
        GKr : float, optional
            Value of GKr, defaults to 0.06
        GKs : float, optional
            Value of GKs, defaults to 0.05
        GNa : float, optional
            Value of GNa, defaults to 35.0
        Gto : float, optional
            Value of Gto, defaults to 0.1
        k_leak : float, optional
            Value of k_leak, defaults to 5e-06
        KmCMDN : float, optional
            Value of KmCMDN, defaults to 2.38
        KmCSQN : float, optional
            Value of KmCSQN, defaults to 800.0
        Kmf : float, optional
            Value of Kmf, defaults to 0.5
        Kmr : float, optional
            Value of Kmr
        KmTRPN : float, optional
            Value of KmTRPN, defaults to 0.5
        kNCX : float, optional
            Value of kNCX, defaults to 2.268e-16
        kRyR : float, optional
            Value of kRyR, defaults to 0.01
        rSL : float, optional
            Value of rSL, defaults to 10.5
        rSR : float, optional
            Value of rSR, defaults to 6.0
        TRPNtot : float, optional
            Value of TRPNtot, defaults to 35.0
        V_sub_SL : float, optional
            Value of V_sub_SL
        V_sub_SR : float, optional
            Value of V_sub_SR
        Vmax : float, optional
            Value of Vmax, defaults to 0.9996
    """
    SHORTNAME = 'rat_neon'
    MODELID = 'Rat_neon'
    PARAMETERS = {'AcapCm2F': None,  'AcapCmFVmyo': None,  'CMDNtot': 50.0,  
                  'CSQNtot': 24750.0,  'DCa': 7.0,  'dx': 0.1,  'GCaL': None,  
                  'GCaT': 0.2,  'Gf': 0.021,  'GK1': 0.0515,  'GKr': 0.06,  
                  'GKs': 0.05,  'GNa': 35.0,  'Gto': 0.1,  'k_leak': 5e-06,  
                  'KmCMDN': 2.38,  'KmCSQN': 800.0,  'Kmf': 0.5,  'Kmr': None,  
                  'KmTRPN': 0.5,  'kNCX': 2.268e-16,  'kRyR': 0.01,  
                  'rSL': 10.5,  'rSR': 6.0,  'TRPNtot': 35.0,  
                  'V_sub_SL': None,  'V_sub_SR': None,  'Vmax': 0.9996}


class RNCIonicModel(_AbstractIonicModel):
    """
    Describes the RNC ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        extern_Cab : float, optional
            Value of extern_Cab
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.24
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.06984
        g_Na : float, optional
            Value of g_Na, defaults to 7.8
        g_to : float, optional
            Value of g_to, defaults to 0.19824
        maxCa_up : float, optional
            Value of maxCa_up, defaults to 15.0
        tau_tr : float, optional
            Value of tau_tr, defaults to 180.0
        tau_u : float, optional
            Value of tau_u, defaults to 8.0
    """
    SHORTNAME = 'rnc'
    MODELID = 'RNC'
    PARAMETERS = {'extern_Cab': None,  'g_CaL': 0.24,  'g_Kr': 0.06984,  
                  'g_Na': 7.8,  'g_to': 0.19824,  'maxCa_up': 15.0,  
                  'tau_tr': 180.0,  'tau_u': 8.0}


class SCIonicModel(_AbstractIonicModel):
    """
    Describes the SC ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        k : float, optional
            Value of k, defaults to 0.0
        wKr : float, optional
            Value of wKr, defaults to 0.0
        wKs : float, optional
            Value of wKs, defaults to 0.0
    """
    SHORTNAME = 'sc'
    MODELID = 'SC'
    PARAMETERS = {'k': 0.0,  'wKr': 0.0,  'wKs': 0.0}


class TT2IonicModel(_AbstractIonicModel):
    """
    Describes the TT2 ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        big_cnt : float, optional
            Value of big_cnt, defaults to 0.0
        Bufc : float, optional
            Value of Bufc, defaults to 0.2
        Bufsr : float, optional
            Value of Bufsr, defaults to 10.0
        Cao : float, optional
            Value of Cao, defaults to 2.0
        CAPACITANCE : float, optional
            Value of CAPACITANCE, defaults to 0.185
        cell_type : float, optional
            Value of cell_type, defaults to 0.0
        F : float, optional
            Value of F, defaults to 96485.3415
        flags : float, optional
            Value of flags
        fmode : float, optional
            Value of fmode, defaults to 0.0
        GbCa : float, optional
            Value of GbCa, defaults to 0.000592
        GbNa : float, optional
            Value of GbNa, defaults to 0.00029
        GCaL : float, optional
            Value of GCaL, defaults to 3.98e-05
        GK1 : float, optional
            Value of GK1, defaults to 5.405
        Gkr : float, optional
            Value of Gkr, defaults to 0.153
        Gks : float, optional
            Value of Gks, defaults to 0.392
        GNa : float, optional
            Value of GNa, defaults to 14.838
        GpCa : float, optional
            Value of GpCa, defaults to 0.1238
        GpK : float, optional
            Value of GpK, defaults to 0.0146
        Gto : float, optional
            Value of Gto, defaults to 0.073
        IpK : float, optional
            Value of IpK
        Kbufc : float, optional
            Value of Kbufc, defaults to 0.001
        Kbufsr : float, optional
            Value of Kbufsr, defaults to 0.3
        KmCa : float, optional
            Value of KmCa, defaults to 1.38
        KmK : float, optional
            Value of KmK, defaults to 1.0
        KmNa : float, optional
            Value of KmNa, defaults to 40.0
        KmNai : float, optional
            Value of KmNai, defaults to 87.5
        knaca : float, optional
            Value of knaca, defaults to 1000.0
        knak : float, optional
            Value of knak, defaults to 2.724
        Ko : float, optional
            Value of Ko, defaults to 5.4
        KpCa : float, optional
            Value of KpCa, defaults to 0.0005
        ksat : float, optional
            Value of ksat, defaults to 0.1
        Kup : float, optional
            Value of Kup, defaults to 0.00025
        n : float, optional
            Value of n, defaults to 0.35
        Nao : float, optional
            Value of Nao, defaults to 140.0
        pKNa : float, optional
            Value of pKNa, defaults to 0.03
        R : float, optional
            Value of R, defaults to 8314.472
        RTONF : float, optional
            Value of RTONF
        scl_tau_f : float, optional
            Value of scl_tau_f, defaults to 1.0
        T : float, optional
            Value of T, defaults to 310.0
        Vc : float, optional
            Value of Vc, defaults to 0.016404
        Vmaxup : float, optional
            Value of Vmaxup, defaults to 0.006375
        Vsr : float, optional
            Value of Vsr, defaults to 0.001094
    """
    SHORTNAME = 'tt2'
    MODELID = 'TT2'
    PARAMETERS = {'big_cnt': 0.0,  'Bufc': 0.2,  'Bufsr': 10.0,  'Cao': 2.0,  
                  'CAPACITANCE': 0.185,  'cell_type': 0.0,  'F': 96485.3415,  
                  'flags': None,  'fmode': 0.0,  'GbCa': 0.000592,  
                  'GbNa': 0.00029,  'GCaL': 3.98e-05,  'GK1': 5.405,  
                  'Gkr': 0.153,  'Gks': 0.392,  'GNa': 14.838,  'GpCa': 0.1238,  
                  'GpK': 0.0146,  'Gto': 0.073,  'IpK': None,  'Kbufc': 0.001,  
                  'Kbufsr': 0.3,  'KmCa': 1.38,  'KmK': 1.0,  'KmNa': 40.0,  
                  'KmNai': 87.5,  'knaca': 1000.0,  'knak': 2.724,  'Ko': 5.4,  
                  'KpCa': 0.0005,  'ksat': 0.1,  'Kup': 0.00025,  'n': 0.35,  
                  'Nao': 140.0,  'pKNa': 0.03,  'R': 8314.472,  'RTONF': None,  
                  'scl_tau_f': 1.0,  'T': 310.0,  'Vc': 0.016404,  
                  'Vmaxup': 0.006375,  'Vsr': 0.001094}


class TTIonicModel(_AbstractIonicModel):
    """
    Describes the TT ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Bufc : float, optional
            Value of Bufc, defaults to 0.15
        Bufsr : float, optional
            Value of Bufsr, defaults to 10.0
        Cao : float, optional
            Value of Cao, defaults to 2.0
        CAPACITANCE : float, optional
            Value of CAPACITANCE, defaults to 0.185
        cell_type : float, optional
            Value of cell_type, defaults to 0.0
        F : float, optional
            Value of F, defaults to 96485.3415
        flags : float, optional
            Value of flags
        fmode : float, optional
            Value of fmode
        GbCa : float, optional
            Value of GbCa, defaults to 0.000592
        GbNa : float, optional
            Value of GbNa, defaults to 0.00029
        GCaL : float, optional
            Value of GCaL, defaults to 0.000175
        GK1 : float, optional
            Value of GK1, defaults to 5.405
        Gkr : float, optional
            Value of Gkr, defaults to 0.096
        Gks : float, optional
            Value of Gks, defaults to 0.245
        GNa : float, optional
            Value of GNa, defaults to 14.838
        GpCa : float, optional
            Value of GpCa, defaults to 0.825
        GpK : float, optional
            Value of GpK, defaults to 0.0146
        Gto : float, optional
            Value of Gto, defaults to 0.073
        IpK : float, optional
            Value of IpK
        Kbufc : float, optional
            Value of Kbufc, defaults to 0.001
        Kbufsr : float, optional
            Value of Kbufsr, defaults to 0.3
        KmCa : float, optional
            Value of KmCa, defaults to 1.38
        KmK : float, optional
            Value of KmK, defaults to 1.0
        KmNa : float, optional
            Value of KmNa, defaults to 40.0
        KmNai : float, optional
            Value of KmNai, defaults to 87.5
        knaca : float, optional
            Value of knaca, defaults to 1000.0
        knak : float, optional
            Value of knak, defaults to 1.362
        Ko : float, optional
            Value of Ko, defaults to 5.4
        KpCa : float, optional
            Value of KpCa, defaults to 0.0005
        ksat : float, optional
            Value of ksat, defaults to 0.1
        Kup : float, optional
            Value of Kup, defaults to 0.00025
        n : float, optional
            Value of n, defaults to 0.35
        Nao : float, optional
            Value of Nao, defaults to 140.0
        pKNa : float, optional
            Value of pKNa, defaults to 0.03
        R : float, optional
            Value of R, defaults to 8314.472
        RTONF : float, optional
            Value of RTONF
        T : float, optional
            Value of T, defaults to 310.0
        taufca : float, optional
            Value of taufca, defaults to 2.0
        taug : float, optional
            Value of taug, defaults to 2.0
        Vc : float, optional
            Value of Vc, defaults to 0.016404
        Vmaxup : float, optional
            Value of Vmaxup, defaults to 0.000425
        Vsr : float, optional
            Value of Vsr, defaults to 0.001094
    """
    SHORTNAME = 'tt'
    MODELID = 'TT'
    PARAMETERS = {'Bufc': 0.15,  'Bufsr': 10.0,  'Cao': 2.0,  
                  'CAPACITANCE': 0.185,  'cell_type': 0.0,  'F': 96485.3415,  
                  'flags': None,  'fmode': None,  'GbCa': 0.000592,  
                  'GbNa': 0.00029,  'GCaL': 0.000175,  'GK1': 5.405,  
                  'Gkr': 0.096,  'Gks': 0.245,  'GNa': 14.838,  'GpCa': 0.825,  
                  'GpK': 0.0146,  'Gto': 0.073,  'IpK': None,  'Kbufc': 0.001,  
                  'Kbufsr': 0.3,  'KmCa': 1.38,  'KmK': 1.0,  'KmNa': 40.0,  
                  'KmNai': 87.5,  'knaca': 1000.0,  'knak': 1.362,  'Ko': 5.4,  
                  'KpCa': 0.0005,  'ksat': 0.1,  'Kup': 0.00025,  'n': 0.35,  
                  'Nao': 140.0,  'pKNa': 0.03,  'R': 8314.472,  'RTONF': None,  
                  'T': 310.0,  'taufca': 2.0,  'taug': 2.0,  'Vc': 0.016404,  
                  'Vmaxup': 0.000425,  'Vsr': 0.001094}


class TTREDIonicModel(_AbstractIonicModel):
    """
    Describes the TTRED ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Bufc : float, optional
            Value of Bufc
        Bufsr : float, optional
            Value of Bufsr
        Cao : float, optional
            Value of Cao, defaults to 2.0
        CAPACITANCE : float, optional
            Value of CAPACITANCE, defaults to 0.185
        cell_type : float, optional
            Value of cell_type
        fmode : float, optional
            Value of fmode
        GbCa : float, optional
            Value of GbCa, defaults to 0.000592
        GbNa : float, optional
            Value of GbNa, defaults to 0.00029
        GCaL : float, optional
            Value of GCaL, defaults to 0.2786
        GK1 : float, optional
            Value of GK1, defaults to 5.405
        Gkr : float, optional
            Value of Gkr, defaults to 0.101
        Gks : float, optional
            Value of Gks, defaults to 0.392
        GNa : float, optional
            Value of GNa, defaults to 14.838
        GpCa : float, optional
            Value of GpCa, defaults to 0.1238
        GpK : float, optional
            Value of GpK, defaults to 0.0293
        Gto : float, optional
            Value of Gto, defaults to 0.073
        IpK : float, optional
            Value of IpK
        Kbufc : float, optional
            Value of Kbufc
        Kbufsr : float, optional
            Value of Kbufsr
        KmCa : float, optional
            Value of KmCa, defaults to 1.38
        KmK : float, optional
            Value of KmK, defaults to 1.0
        KmNa : float, optional
            Value of KmNa, defaults to 40.0
        KmNai : float, optional
            Value of KmNai, defaults to 87.5
        knaca : float, optional
            Value of knaca, defaults to 1000.0
        knak : float, optional
            Value of knak, defaults to 2.724
        Ko : float, optional
            Value of Ko, defaults to 5.4
        KpCa : float, optional
            Value of KpCa, defaults to 0.0005
        ksat : float, optional
            Value of ksat, defaults to 0.1
        Kup : float, optional
            Value of Kup
        n : float, optional
            Value of n, defaults to 0.35
        Nao : float, optional
            Value of Nao, defaults to 140.0
        pKNa : float, optional
            Value of pKNa, defaults to 0.03
        RTONF : float, optional
            Value of RTONF
        T : float, optional
            Value of T, defaults to 310.0
        Vc : float, optional
            Value of Vc, defaults to 0.016404
        Vmaxup : float, optional
            Value of Vmaxup
        Vsr : float, optional
            Value of Vsr, defaults to 0.001094
    """
    SHORTNAME = 'ttred'
    MODELID = 'TTRED'
    PARAMETERS = {'Bufc': None,  'Bufsr': None,  'Cao': 2.0,  
                  'CAPACITANCE': 0.185,  'cell_type': None,  'fmode': None,  
                  'GbCa': 0.000592,  'GbNa': 0.00029,  'GCaL': 0.2786,  
                  'GK1': 5.405,  'Gkr': 0.101,  'Gks': 0.392,  'GNa': 14.838,  
                  'GpCa': 0.1238,  'GpK': 0.0293,  'Gto': 0.073,  'IpK': None,  
                  'Kbufc': None,  'Kbufsr': None,  'KmCa': 1.38,  'KmK': 1.0,  
                  'KmNa': 40.0,  'KmNai': 87.5,  'knaca': 1000.0,  
                  'knak': 2.724,  'Ko': 5.4,  'KpCa': 0.0005,  'ksat': 0.1,  
                  'Kup': None,  'n': 0.35,  'Nao': 140.0,  'pKNa': 0.03,  
                  'RTONF': None,  'T': 310.0,  'Vc': 0.016404,  'Vmaxup': None,  
                  'Vsr': 0.001094}


class UCLA_RABIonicModel(_AbstractIonicModel):
    """
    Describes the UCLA_RAB ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        blf_ICaL : float, optional
            Value of blf_ICaL, defaults to 1.0
        blf_IKr : float, optional
            Value of blf_IKr, defaults to 1.0
        blf_IKs : float, optional
            Value of blf_IKs, defaults to 1.0
        blf_INa : float, optional
            Value of blf_INa, defaults to 1.0
        blf_INaCa : float, optional
            Value of blf_INaCa, defaults to 1.0
        blf_recov : float, optional
            Value of blf_recov, defaults to 1.0
        blf_tau_h : float, optional
            Value of blf_tau_h, defaults to 1.0
        Cao : float, optional
            Value of Cao, defaults to 1.8
        d_g : float, optional
            Value of d_g, defaults to 2.58079
        d_taur : float, optional
            Value of d_taur, defaults to 30.0
        gCa : float, optional
            Value of gCa, defaults to 182.0
        gK1 : float, optional
            Value of gK1, defaults to 0.3
        gKr : float, optional
            Value of gKr, defaults to 0.0125
        gKs : float, optional
            Value of gKs, defaults to 0.32
        gNa : float, optional
            Value of gNa, defaults to 12.0
        gNaK : float, optional
            Value of gNaK, defaults to 1.5
        gss : float, optional
            Value of gss
        gtof : float, optional
            Value of gtof, defaults to 0.11
        gtos : float, optional
            Value of gtos, defaults to 0.04
        K_i : float, optional
            Value of K_i, defaults to 140.0
        Ko : float, optional
            Value of Ko, defaults to 5.4
        Na_i : float, optional
            Value of Na_i, defaults to 14.01807252
        Nao : float, optional
            Value of Nao, defaults to 136.0
        T : float, optional
            Value of T, defaults to 308.0
        Vup : float, optional
            Value of Vup, defaults to 0.4
    """
    SHORTNAME = 'ucla_rab'
    MODELID = 'UCLA_RAB'
    PARAMETERS = {'blf_ICaL': 1.0,  'blf_IKr': 1.0,  'blf_IKs': 1.0,  
                  'blf_INa': 1.0,  'blf_INaCa': 1.0,  'blf_recov': 1.0,  
                  'blf_tau_h': 1.0,  'Cao': 1.8,  'd_g': 2.58079,  
                  'd_taur': 30.0,  'gCa': 182.0,  'gK1': 0.3,  'gKr': 0.0125,  
                  'gKs': 0.32,  'gNa': 12.0,  'gNaK': 1.5,  'gss': None,  
                  'gtof': 0.11,  'gtos': 0.04,  'K_i': 140.0,  'Ko': 5.4,  
                  'Na_i': 14.01807252,  'Nao': 136.0,  'T': 308.0,  'Vup': 0.4}


class WANG_NNRIonicModel(_AbstractIonicModel):
    """
    Describes the WANG_NNR ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        Ca_i_init : float, optional
            Value of Ca_i_init, defaults to 0.2049
        i_CaL_slowdown : float, optional
            Value of i_CaL_slowdown, defaults to 1.0
        Ki_init : float, optional
            Value of Ki_init, defaults to 136450.0
        Nai_init : float, optional
            Value of Nai_init, defaults to 21747.0
    """
    SHORTNAME = 'wang_nnr'
    MODELID = 'WANG_NNR'
    PARAMETERS = {'Ca_i_init': 0.2049,  'i_CaL_slowdown': 1.0,  
                  'Ki_init': 136450.0,  'Nai_init': 21747.0}


class SHANNONIonicModel(_AbstractIonicModel):
    """
    Describes the SHANNON ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
    """
    SHORTNAME = 'shannon'
    MODELID = 'SHANNON'
    PARAMETERS = {}


class koivumakiIonicModel(_AbstractIonicModel):
    """
    Describes the koivumaki ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        base_phos : float, optional
            Value of base_phos, defaults to 0.1
        cAF_cpumps : float, optional
            Value of cAF_cpumps, defaults to 1.0
        cAF_lcell : float, optional
            Value of cAF_lcell, defaults to 1.0
        cAF_RyR : float, optional
            Value of cAF_RyR, defaults to 1.0
        g_B_Ca : float, optional
            Value of g_B_Ca, defaults to 8.4e-05
        g_B_Na : float, optional
            Value of g_B_Na, defaults to 6.0599e-05
        g_Ca_L : float, optional
            Value of g_Ca_L, defaults to 0.007
        g_K1 : float, optional
            Value of g_K1, defaults to 0.0029
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0034
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.000175
        g_sus : float, optional
            Value of g_sus, defaults to 0.00225
        g_t : float, optional
            Value of g_t, defaults to 0.0011
        gNa : float, optional
            Value of gNa, defaults to 0.558
        gNaL : float, optional
            Value of gNaL, defaults to 0.000423
        k_NaCa : float, optional
            Value of k_NaCa, defaults to 0.008433945
        PLB_SERCA_ratio : float, optional
            Value of PLB_SERCA_ratio, defaults to 1.0
        SLN_SERCA_ratio : float, optional
            Value of SLN_SERCA_ratio, defaults to 1.0
    """
    SHORTNAME = 'koivumaki'
    MODELID = 'koivumaki'
    PARAMETERS = {'base_phos': 0.1,  'cAF_cpumps': 1.0,  'cAF_lcell': 1.0,  
                  'cAF_RyR': 1.0,  'g_B_Ca': 8.4e-05,  'g_B_Na': 6.0599e-05,  
                  'g_Ca_L': 0.007,  'g_K1': 0.0029,  'g_Kr': 0.0034,  
                  'g_Ks': 0.000175,  'g_sus': 0.00225,  'g_t': 0.0011,  
                  'gNa': 0.558,  'gNaL': 0.000423,  'k_NaCa': 0.008433945,  
                  'PLB_SERCA_ratio': 1.0,  'SLN_SERCA_ratio': 1.0}


class koivumaki_newINaIonicModel(_AbstractIonicModel):
    """
    Describes the koivumaki_newINa ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        cAF_cpumps : float, optional
            Value of cAF_cpumps
        cAF_g_f : float, optional
            Value of cAF_g_f, defaults to 1.0
        cAF_gCaL : float, optional
            Value of cAF_gCaL
        cAF_gK1 : float, optional
            Value of cAF_gK1
        cAF_gKCa : float, optional
            Value of cAF_gKCa
        cAF_gKs : float, optional
            Value of cAF_gKs
        cAF_gNa : float, optional
            Value of cAF_gNa
        cAF_gsus : float, optional
            Value of cAF_gsus
        cAF_gt : float, optional
            Value of cAF_gt
        cAF_kNaCa : float, optional
            Value of cAF_kNaCa
        cAF_lcell : float, optional
            Value of cAF_lcell
        cAF_phos : float, optional
            Value of cAF_phos
        cAF_PLB : float, optional
            Value of cAF_PLB
        cAF_RyR : float, optional
            Value of cAF_RyR
        cAF_SLN : float, optional
            Value of cAF_SLN
        flags : float, optional
            Value of flags
        mode : float, optional
            Value of mode
    """
    SHORTNAME = 'koivumaki_newina'
    MODELID = 'koivumaki_newINa'
    PARAMETERS = {'cAF_cpumps': None,  'cAF_g_f': 1.0,  'cAF_gCaL': None,  
                  'cAF_gK1': None,  'cAF_gKCa': None,  'cAF_gKs': None,  
                  'cAF_gNa': None,  'cAF_gsus': None,  'cAF_gt': None,  
                  'cAF_kNaCa': None,  'cAF_lcell': None,  'cAF_phos': None,  
                  'cAF_PLB': None,  'cAF_RyR': None,  'cAF_SLN': None,  
                  'flags': None,  'mode': None}


class BOIonicModel(_AbstractIonicModel):
    """
    Describes the BO ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        flags : float, optional
            Value of flags
        k_s : float, optional
            Value of k_s
        k_so : float, optional
            Value of k_so
        k_w_minus : float, optional
            Value of k_w_minus
        mode : float, optional
            Value of mode
        tau_fi : float, optional
            Value of tau_fi
        tau_o1 : float, optional
            Value of tau_o1
        tau_o2 : float, optional
            Value of tau_o2
        tau_s1 : float, optional
            Value of tau_s1
        tau_s2 : float, optional
            Value of tau_s2
        tau_si : float, optional
            Value of tau_si
        tau_so1 : float, optional
            Value of tau_so1
        tau_so2 : float, optional
            Value of tau_so2
        tau_v1_minus : float, optional
            Value of tau_v1_minus
        tau_v2_minus : float, optional
            Value of tau_v2_minus
        tau_v_pls : float, optional
            Value of tau_v_pls, defaults to 1.4506
        tau_w1_minus : float, optional
            Value of tau_w1_minus
        tau_w2_minus : float, optional
            Value of tau_w2_minus
        tau_w_inf : float, optional
            Value of tau_w_inf
        tau_w_pls : float, optional
            Value of tau_w_pls
        th_o : float, optional
            Value of th_o
        th_v : float, optional
            Value of th_v
        th_v_minus : float, optional
            Value of th_v_minus
        th_w : float, optional
            Value of th_w
        u_o : float, optional
            Value of u_o, defaults to 0.0
        u_s : float, optional
            Value of u_s
        u_so : float, optional
            Value of u_so
        u_u : float, optional
            Value of u_u
        u_w_minus : float, optional
            Value of u_w_minus
        w_inf_star : float, optional
            Value of w_inf_star
    """
    SHORTNAME = 'bo'
    MODELID = 'BO'
    PARAMETERS = {'flags': None,  'k_s': None,  'k_so': None,  
                  'k_w_minus': None,  'mode': None,  'tau_fi': None,  
                  'tau_o1': None,  'tau_o2': None,  'tau_s1': None,  
                  'tau_s2': None,  'tau_si': None,  'tau_so1': None,  
                  'tau_so2': None,  'tau_v1_minus': None,  
                  'tau_v2_minus': None,  'tau_v_pls': 1.4506,  
                  'tau_w1_minus': None,  'tau_w2_minus': None,  
                  'tau_w_inf': None,  'tau_w_pls': None,  'th_o': None,  
                  'th_v': None,  'th_v_minus': None,  'th_w': None,  
                  'u_o': 0.0,  'u_s': None,  'u_so': None,  'u_u': None,  
                  'u_w_minus': None,  'w_inf_star': None}


class ToRORd_dynClIonicModel(_AbstractIonicModel):
    """
    Describes the ToRORd_dynCl ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        A_atp : float, optional
            Value of A_atp, defaults to 2.0
        aCaMK : float, optional
            Value of aCaMK, defaults to 0.05
        Aff : float, optional
            Value of Aff, defaults to 0.6
        alpha_1 : float, optional
            Value of alpha_1, defaults to 0.154375
        bCaMK : float, optional
            Value of bCaMK, defaults to 0.00068
        beta_1 : float, optional
            Value of beta_1, defaults to 0.1911
        BSLmax : float, optional
            Value of BSLmax, defaults to 1.124
        BSRmax : float, optional
            Value of BSRmax, defaults to 0.047
        bt : float, optional
            Value of bt, defaults to 4.75
        cajsr_half : float, optional
            Value of cajsr_half, defaults to 1.7
        CaMKo : float, optional
            Value of CaMKo, defaults to 0.05
        cao : float, optional
            Value of cao, defaults to 1.8
        celltype : float, optional
            Value of celltype, defaults to 0.0
        clo : float, optional
            Value of clo, defaults to 150.0
        cmdnmax_b : float, optional
            Value of cmdnmax_b, defaults to 0.05
        csqnmax : float, optional
            Value of csqnmax, defaults to 10.0
        delta : float, optional
            Value of delta, defaults to -0.155
        dielConstant : float, optional
            Value of dielConstant, defaults to 74.0
        EKshift : float, optional
            Value of EKshift, defaults to 0.0
        eP : float, optional
            Value of eP, defaults to 4.2
        F : float, optional
            Value of F, defaults to 96485.0
        Fjunc : float, optional
            Value of Fjunc, defaults to 1.0
        fkatp : float, optional
            Value of fkatp, defaults to 0.0
        flags : float, optional
            Value of flags
        GClb : float, optional
            Value of GClb, defaults to 0.00198
        GClCa : float, optional
            Value of GClCa, defaults to 0.2843
        GK1_b : float, optional
            Value of GK1_b, defaults to 0.6992
        gkatp : float, optional
            Value of gkatp, defaults to 4.3195
        GKb_b : float, optional
            Value of GKb_b, defaults to 0.0189
        GKr_b : float, optional
            Value of GKr_b, defaults to 0.0321
        GKs_b : float, optional
            Value of GKs_b, defaults to 0.0011
        GNa : float, optional
            Value of GNa, defaults to 11.7802
        GNaL_b : float, optional
            Value of GNaL_b, defaults to 0.0279
        Gncx_b : float, optional
            Value of Gncx_b, defaults to 0.0034
        GpCa : float, optional
            Value of GpCa, defaults to 0.0005
        Gto_b : float, optional
            Value of Gto_b, defaults to 0.16
        H : float, optional
            Value of H, defaults to 1e-07
        ICaL_fractionSS : float, optional
            Value of ICaL_fractionSS, defaults to 0.8
        INaCa_fractionSS : float, optional
            Value of INaCa_fractionSS, defaults to 0.35
        Jrel_b : float, optional
            Value of Jrel_b, defaults to 1.5378
        Jup_b : float, optional
            Value of Jup_b, defaults to 1.0
        k1m : float, optional
            Value of k1m, defaults to 182.4
        k1p : float, optional
            Value of k1p, defaults to 949.5
        k2m : float, optional
            Value of k2m, defaults to 39.4
        k2n : float, optional
            Value of k2n, defaults to 500.0
        k2p : float, optional
            Value of k2p, defaults to 687.2
        k3m : float, optional
            Value of k3m, defaults to 79300.0
        k3p : float, optional
            Value of k3p, defaults to 1899.0
        k4m : float, optional
            Value of k4m, defaults to 40.0
        k4p : float, optional
            Value of k4p, defaults to 639.0
        K_atp : float, optional
            Value of K_atp, defaults to 0.25
        K_o_n : float, optional
            Value of K_o_n, defaults to 5.0
        kasymm : float, optional
            Value of kasymm, defaults to 12.5
        kcaoff : float, optional
            Value of kcaoff, defaults to 5000.0
        kcaon : float, optional
            Value of kcaon, defaults to 1500000.0
        KdClCa : float, optional
            Value of KdClCa, defaults to 0.1
        Khp : float, optional
            Value of Khp, defaults to 1.698e-07
        Kki : float, optional
            Value of Kki, defaults to 0.5
        Kko : float, optional
            Value of Kko, defaults to 0.3582
        KmBSL : float, optional
            Value of KmBSL, defaults to 0.0087
        KmBSR : float, optional
            Value of KmBSR, defaults to 0.00087
        KmCaAct : float, optional
            Value of KmCaAct, defaults to 0.00015
        KmCaM : float, optional
            Value of KmCaM, defaults to 0.0015
        KmCaMK : float, optional
            Value of KmCaMK, defaults to 0.15
        KmCap : float, optional
            Value of KmCap, defaults to 0.0005
        kmcmdn : float, optional
            Value of kmcmdn, defaults to 0.00238
        kmcsqn : float, optional
            Value of kmcsqn, defaults to 0.8
        Kmgatp : float, optional
            Value of Kmgatp, defaults to 1.698e-07
        Kmn : float, optional
            Value of Kmn, defaults to 0.002
        kmtrpn : float, optional
            Value of kmtrpn, defaults to 0.0005
        kna1 : float, optional
            Value of kna1, defaults to 15.0
        kna2 : float, optional
            Value of kna2, defaults to 5.0
        kna3 : float, optional
            Value of kna3, defaults to 88.12
        Knai0 : float, optional
            Value of Knai0, defaults to 9.073
        Knao0 : float, optional
            Value of Knao0, defaults to 27.78
        Knap : float, optional
            Value of Knap, defaults to 224.0
        ko : float, optional
            Value of ko, defaults to 5.0
        Kxkur : float, optional
            Value of Kxkur, defaults to 292.0
        L : float, optional
            Value of L, defaults to 0.01
        MgADP : float, optional
            Value of MgADP, defaults to 0.05
        MgATP : float, optional
            Value of MgATP, defaults to 9.8
        nao : float, optional
            Value of nao, defaults to 140.0
        offset : float, optional
            Value of offset, defaults to 0.0
        PCa_b : float, optional
            Value of PCa_b, defaults to 8.3757e-05
        PCab : float, optional
            Value of PCab, defaults to 5.9194e-08
        PKNa : float, optional
            Value of PKNa, defaults to 0.01833
        PNab : float, optional
            Value of PNab, defaults to 1.9239e-09
        Pnak_b : float, optional
            Value of Pnak_b, defaults to 15.4509
        qca : float, optional
            Value of qca, defaults to 0.167
        qna : float, optional
            Value of qna, defaults to 0.5224
        R : float, optional
            Value of R, defaults to 8314.0
        rad : float, optional
            Value of rad, defaults to 0.0011
        T : float, optional
            Value of T, defaults to 310.0
        tauCa : float, optional
            Value of tauCa, defaults to 0.2
        tauCl : float, optional
            Value of tauCl, defaults to 2.0
        tauK : float, optional
            Value of tauK, defaults to 2.0
        tauNa : float, optional
            Value of tauNa, defaults to 2.0
        thL : float, optional
            Value of thL, defaults to 200.0
        tjca : float, optional
            Value of tjca, defaults to 72.5
        trpnmax : float, optional
            Value of trpnmax, defaults to 0.07
        vShift : float, optional
            Value of vShift, defaults to 0.0
        wca : float, optional
            Value of wca, defaults to 60000.0
        wna : float, optional
            Value of wna, defaults to 60000.0
        wnaca : float, optional
            Value of wnaca, defaults to 5000.0
        zca : float, optional
            Value of zca, defaults to 2.0
        zcl : float, optional
            Value of zcl, defaults to -1.0
        zk : float, optional
            Value of zk, defaults to 1.0
        zna : float, optional
            Value of zna, defaults to 1.0
    """
    SHORTNAME = 'torord_dyncl'
    MODELID = 'ToRORd_dynCl'
    PARAMETERS = {'A_atp': 2.0,  'aCaMK': 0.05,  'Aff': 0.6,  
                  'alpha_1': 0.154375,  'bCaMK': 0.00068,  'beta_1': 0.1911,  
                  'BSLmax': 1.124,  'BSRmax': 0.047,  'bt': 4.75,  
                  'cajsr_half': 1.7,  'CaMKo': 0.05,  'cao': 1.8,  
                  'celltype': 0.0,  'clo': 150.0,  'cmdnmax_b': 0.05,  
                  'csqnmax': 10.0,  'delta': -0.155,  'dielConstant': 74.0,  
                  'EKshift': 0.0,  'eP': 4.2,  'F': 96485.0,  'Fjunc': 1.0,  
                  'fkatp': 0.0,  'flags': None,  'GClb': 0.00198,  
                  'GClCa': 0.2843,  'GK1_b': 0.6992,  'gkatp': 4.3195,  
                  'GKb_b': 0.0189,  'GKr_b': 0.0321,  'GKs_b': 0.0011,  
                  'GNa': 11.7802,  'GNaL_b': 0.0279,  'Gncx_b': 0.0034,  
                  'GpCa': 0.0005,  'Gto_b': 0.16,  'H': 1e-07,  
                  'ICaL_fractionSS': 0.8,  'INaCa_fractionSS': 0.35,  
                  'Jrel_b': 1.5378,  'Jup_b': 1.0,  'k1m': 182.4,  
                  'k1p': 949.5,  'k2m': 39.4,  'k2n': 500.0,  'k2p': 687.2,  
                  'k3m': 79300.0,  'k3p': 1899.0,  'k4m': 40.0,  'k4p': 639.0,  
                  'K_atp': 0.25,  'K_o_n': 5.0,  'kasymm': 12.5,  
                  'kcaoff': 5000.0,  'kcaon': 1500000.0,  'KdClCa': 0.1,  
                  'Khp': 1.698e-07,  'Kki': 0.5,  'Kko': 0.3582,  
                  'KmBSL': 0.0087,  'KmBSR': 0.00087,  'KmCaAct': 0.00015,  
                  'KmCaM': 0.0015,  'KmCaMK': 0.15,  'KmCap': 0.0005,  
                  'kmcmdn': 0.00238,  'kmcsqn': 0.8,  'Kmgatp': 1.698e-07,  
                  'Kmn': 0.002,  'kmtrpn': 0.0005,  'kna1': 15.0,  'kna2': 5.0,  
                  'kna3': 88.12,  'Knai0': 9.073,  'Knao0': 27.78,  
                  'Knap': 224.0,  'ko': 5.0,  'Kxkur': 292.0,  'L': 0.01,  
                  'MgADP': 0.05,  'MgATP': 9.8,  'nao': 140.0,  'offset': 0.0,  
                  'PCa_b': 8.3757e-05,  'PCab': 5.9194e-08,  'PKNa': 0.01833,  
                  'PNab': 1.9239e-09,  'Pnak_b': 15.4509,  'qca': 0.167,  
                  'qna': 0.5224,  'R': 8314.0,  'rad': 0.0011,  'T': 310.0,  
                  'tauCa': 0.2,  'tauCl': 2.0,  'tauK': 2.0,  'tauNa': 2.0,  
                  'thL': 200.0,  'tjca': 72.5,  'trpnmax': 0.07,  
                  'vShift': 0.0,  'wca': 60000.0,  'wna': 60000.0,  
                  'wnaca': 5000.0,  'zca': 2.0,  'zcl': -1.0,  'zk': 1.0,  
                  'zna': 1.0}


class GTT2_fastIonicModel(_AbstractIonicModel):
    """
    Describes the GTT2_fast ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        ENDO : float, optional
            Value of ENDO, defaults to 0.0
        G_bCa : float, optional
            Value of G_bCa, defaults to 0.000592
        G_bNa : float, optional
            Value of G_bNa, defaults to 0.00029
        G_CaL : float, optional
            Value of G_CaL, defaults to 3.98e-05
        G_K1 : float, optional
            Value of G_K1, defaults to 5.405
        G_Kr : float, optional
            Value of G_Kr, defaults to 0.153
        G_Ks : float, optional
            Value of G_Ks, defaults to 0.392
        G_leak : float, optional
            Value of G_leak, defaults to 0.00036
        G_Na : float, optional
            Value of G_Na, defaults to 14.838
        G_NaCa : float, optional
            Value of G_NaCa, defaults to 1000.0
        G_NaK : float, optional
            Value of G_NaK, defaults to 2.724
        G_NaL : float, optional
            Value of G_NaL, defaults to 0.0075
        G_pCa : float, optional
            Value of G_pCa, defaults to 0.1238
        G_pK : float, optional
            Value of G_pK, defaults to 0.0146
        G_rel : float, optional
            Value of G_rel, defaults to 0.102
        G_tf : float, optional
            Value of G_tf, defaults to 1.0
        G_to : float, optional
            Value of G_to, defaults to 0.294
        G_up : float, optional
            Value of G_up, defaults to 0.006375
        G_vxfer : float, optional
            Value of G_vxfer, defaults to 0.0038
    """
    SHORTNAME = 'gtt2_fast'
    MODELID = 'GTT2_fast'
    PARAMETERS = {'ENDO': 0.0,  'G_bCa': 0.000592,  'G_bNa': 0.00029,  
                  'G_CaL': 3.98e-05,  'G_K1': 5.405,  'G_Kr': 0.153,  
                  'G_Ks': 0.392,  'G_leak': 0.00036,  'G_Na': 14.838,  
                  'G_NaCa': 1000.0,  'G_NaK': 2.724,  'G_NaL': 0.0075,  
                  'G_pCa': 0.1238,  'G_pK': 0.0146,  'G_rel': 0.102,  
                  'G_tf': 1.0,  'G_to': 0.294,  'G_up': 0.006375,  
                  'G_vxfer': 0.0038}


class GTT2_fast_PURKIonicModel(_AbstractIonicModel):
    """
    Describes the GTT2_fast_PURK ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        D_CaL_off : float, optional
            Value of D_CaL_off, defaults to 0.0
        ENDO : float, optional
            Value of ENDO, defaults to 0.0
        G_bCa : float, optional
            Value of G_bCa, defaults to 0.000592
        G_bNa : float, optional
            Value of G_bNa, defaults to 0.00029
        G_CaL : float, optional
            Value of G_CaL, defaults to 3.98e-05
        G_K1 : float, optional
            Value of G_K1, defaults to 5.405
        G_Kr : float, optional
            Value of G_Kr, defaults to 0.153
        G_Ks : float, optional
            Value of G_Ks, defaults to 0.392
        G_leak : float, optional
            Value of G_leak, defaults to 0.00036
        G_Na : float, optional
            Value of G_Na, defaults to 14.838
        G_NaCa : float, optional
            Value of G_NaCa, defaults to 1000.0
        G_NaK : float, optional
            Value of G_NaK, defaults to 2.724
        G_NaL : float, optional
            Value of G_NaL, defaults to 0.0075
        G_pCa : float, optional
            Value of G_pCa, defaults to 0.1238
        G_pK : float, optional
            Value of G_pK, defaults to 0.0146
        G_rel : float, optional
            Value of G_rel, defaults to 0.102
        G_tf : float, optional
            Value of G_tf, defaults to 1.0
        G_to : float, optional
            Value of G_to, defaults to 0.294
        G_up : float, optional
            Value of G_up, defaults to 0.006375
        G_vxfer : float, optional
            Value of G_vxfer, defaults to 0.0038
        xr2_off : float, optional
            Value of xr2_off, defaults to 0.0
    """
    SHORTNAME = 'gtt2_fast_purk'
    MODELID = 'GTT2_fast_PURK'
    PARAMETERS = {'D_CaL_off': 0.0,  'ENDO': 0.0,  'G_bCa': 0.000592,  
                  'G_bNa': 0.00029,  'G_CaL': 3.98e-05,  'G_K1': 5.405,  
                  'G_Kr': 0.153,  'G_Ks': 0.392,  'G_leak': 0.00036,  
                  'G_Na': 14.838,  'G_NaCa': 1000.0,  'G_NaK': 2.724,  
                  'G_NaL': 0.0075,  'G_pCa': 0.1238,  'G_pK': 0.0146,  
                  'G_rel': 0.102,  'G_tf': 1.0,  'G_to': 0.294,  
                  'G_up': 0.006375,  'G_vxfer': 0.0038,  'xr2_off': 0.0}


class JB_COURTEMANCHEIonicModel(_AbstractIonicModel):
    """
    Describes the JB_COURTEMANCHE ionic model

    Args:
        IDs : list of int
            The element IDs to be included in this material region
        name : str
            The name of the model material region
        ACh : float, optional
            Value of ACh, defaults to 1e-06
        factor_dical : float, optional
            Value of factor_dical, defaults to 0.0
        factor_fical : float, optional
            Value of factor_fical, defaults to 0.0
        factor_ical_ft : float, optional
            Value of factor_ical_ft, defaults to 1.0
        factor_ina_c : float, optional
            Value of factor_ina_c, defaults to 0.0
        factor_ina_m : float, optional
            Value of factor_ina_m, defaults to 0.0
        factor_irel : float, optional
            Value of factor_irel, defaults to 1.0
        factor_itr : float, optional
            Value of factor_itr, defaults to 1.0
        factor_iup : float, optional
            Value of factor_iup, defaults to 1.0
        factor_kur : float, optional
            Value of factor_kur, defaults to 1.0
        factor_oa : float, optional
            Value of factor_oa, defaults to 0.0
        factor_tikr : float, optional
            Value of factor_tikr, defaults to 1.0
        g_bCa : float, optional
            Value of g_bCa, defaults to 0.00113
        g_bNa : float, optional
            Value of g_bNa, defaults to 0.000674
        g_CaL : float, optional
            Value of g_CaL, defaults to 0.1238
        g_K1 : float, optional
            Value of g_K1, defaults to 0.09
        g_Kr : float, optional
            Value of g_Kr, defaults to 0.0294
        g_Ks : float, optional
            Value of g_Ks, defaults to 0.129
        g_Na : float, optional
            Value of g_Na, defaults to 7.8
        g_to : float, optional
            Value of g_to, defaults to 0.1652
        K_o : float, optional
            Value of K_o, defaults to 5.4
        maxI_NaCa : float, optional
            Value of maxI_NaCa, defaults to 1600.0
        maxI_NaK : float, optional
            Value of maxI_NaK, defaults to 0.6
        maxI_pCa : float, optional
            Value of maxI_pCa, defaults to 0.275
        maxI_up : float, optional
            Value of maxI_up, defaults to 0.005
        P_ACh : float, optional
            Value of P_ACh, defaults to 1.0
    """
    SHORTNAME = 'jb_courtemanche'
    MODELID = 'JB_COURTEMANCHE'
    PARAMETERS = {'ACh': 1e-06,  'factor_dical': 0.0,  'factor_fical': 0.0,  
                  'factor_ical_ft': 1.0,  'factor_ina_c': 0.0,  
                  'factor_ina_m': 0.0,  'factor_irel': 1.0,  'factor_itr': 1.0,  
                  'factor_iup': 1.0,  'factor_kur': 1.0,  'factor_oa': 0.0,  
                  'factor_tikr': 1.0,  'g_bCa': 0.00113,  'g_bNa': 0.000674,  
                  'g_CaL': 0.1238,  'g_K1': 0.09,  'g_Kr': 0.0294,  
                  'g_Ks': 0.129,  'g_Na': 7.8,  'g_to': 0.1652,  'K_o': 5.4,  
                  'maxI_NaCa': 1600.0,  'maxI_NaK': 0.6,  'maxI_pCa': 0.275,  
                  'maxI_up': 0.005,  'P_ACh': 1.0}


MAPPING = _OrderedDict()
MAPPING['passive']                = PASSIVEIonicModel
MAPPING['alievpanfilov']          = AlievPanfilovIonicModel
MAPPING['arpf']                   = ARPFIonicModel
MAPPING['steward']                = StewardIonicModel
MAPPING['prd2']                   = PRd2IonicModel
MAPPING['converted_courtemanche'] = converted_COURTEMANCHEIonicModel
MAPPING['converted_difrannoble']  = converted_DiFranNobleIonicModel
MAPPING['converted_lrdii_f']      = converted_LRDII_FIonicModel
MAPPING['converted_mbrdr']        = converted_MBRDRIonicModel
MAPPING['converted_nygren']       = converted_NYGRENIonicModel
MAPPING['converted_rnc']          = converted_RNCIonicModel
MAPPING['converted_tt2']          = converted_TT2IonicModel
MAPPING['converted_ucla_rab']     = converted_UCLA_RABIonicModel
MAPPING['ucla_rab_scr']           = UCLA_RAB_SCRIonicModel
MAPPING['courtemanche']           = COURTEMANCHEIonicModel
MAPPING['difrannoble']            = DiFranNobleIonicModel
MAPPING['ecme']                   = ECMEIonicModel
MAPPING['fhn_rmcc']               = FHN_RMcCIonicModel
MAPPING['fox']                    = FOXIonicModel
MAPPING['gpb']                    = GPBIonicModel
MAPPING['gpvatria']               = GPVatriaIonicModel
MAPPING['gpb_land']               = GPB_LandIonicModel
MAPPING['gpb_landhuman']          = GPB_LandHumanIonicModel
MAPPING['gw_can']                 = GW_CANIonicModel
MAPPING['hamr']                   = hAMrIonicModel
MAPPING['hh']                     = HHIonicModel
MAPPING['iion_src']               = IION_SRCIonicModel
MAPPING['inada_avn']              = INADA_AVNIonicModel
MAPPING['iribe']                  = IribeIonicModel
MAPPING['kurata']                 = KurataIonicModel
MAPPING['lmcg']                   = LMCGIonicModel
MAPPING['lr1']                    = LR1IonicModel
MAPPING['lrdii']                  = LRDIIIonicModel
MAPPING['lrdii_f']                = LRDII_FIonicModel
MAPPING['maccannell_fb']          = MacCannell_FbIonicModel
MAPPING['mitchellschaeffer']      = MitchellSchaefferIonicModel
MAPPING['mms']                    = mMSIonicModel
MAPPING['mbrdr']                  = MBRDRIonicModel
MAPPING['murinemouse']            = MurineMouseIonicModel
MAPPING['nygren']                 = NYGRENIonicModel
MAPPING['ord']                    = ORdIonicModel
MAPPING['pandit']                 = PanditIonicModel
MAPPING['pug']                    = PUGIonicModel
MAPPING['rat_neon']               = Rat_neonIonicModel
MAPPING['rnc']                    = RNCIonicModel
MAPPING['sc']                     = SCIonicModel
MAPPING['tt2']                    = TT2IonicModel
MAPPING['tt']                     = TTIonicModel
MAPPING['ttred']                  = TTREDIonicModel
MAPPING['ucla_rab']               = UCLA_RABIonicModel
MAPPING['wang_nnr']               = WANG_NNRIonicModel
MAPPING['shannon']                = SHANNONIonicModel
MAPPING['koivumaki']              = koivumakiIonicModel
MAPPING['koivumaki_newina']       = koivumaki_newINaIonicModel
MAPPING['bo']                     = BOIonicModel
MAPPING['torord_dyncl']           = ToRORd_dynClIonicModel
MAPPING['gtt2_fast']              = GTT2_fastIonicModel
MAPPING['gtt2_fast_purk']         = GTT2_fast_PURKIonicModel
MAPPING['jb_courtemanche']        = JB_COURTEMANCHEIonicModel


def keys():
    """
    Return a list of string keys to be used with :func:`get`
    """
    return MAPPING.keys()


def get(*args, **kwargs):
    """
    Get a class from its key string from :func:`keys`
    """
    return MAPPING.get(*args, **kwargs)


def summary(indent=2, ncolumns=4, padding=2):
    """
    Return a string summarizing the models and their parameters
    """
    modelattr = ['SHORTNAME', 'MODELID']
    maxattrlen = max(map(len, modelattr+['PARAMETERS']))
    summarystr = 'IONIC MODELS\n\n'
    indentstr = ' '*indent
    for name, model in MAPPING.items():
        summarystr += indentstr+'{}\n'.format(name)
        for attr in modelattr:
            obj = getattr(model, attr, None)
            if obj is None:
                continue
            summarystr += indentstr*2+'{}: {}\n'.format(attr.lower().ljust(maxattrlen), str(obj))
        attr = 'PARAMETERS'
        obj = getattr(model, attr, None)
        if obj is not None and len(obj) > 0:
            tablestr = stringtable(obj, ncolumns, padding, rowprefix=' '*(maxattrlen+2*indent+2))
            summarystr += indentstr*2+attr.lower().ljust(maxattrlen)+': '+tablestr[len(attr)+2*indent+2:]
        summarystr += '\n'
    return summarystr
