#!/usr/bin/env python

"""
Overview
========

.. code-block:: bash

    cd ${TUTORIALS}/02_EP_tissue/01_basic_usage

This example introduces to the basics of using the openCARP executable 
for simulating EP at the tissue and organ scale.
Details on how to inquire input parameters are provided in the :ref:`openCARP <openCARP>` section
of the manual.

In absence of pacemaker currents electrical stimulus currents must be supplied in order to 
depolarize tissue above the threshold to elicit a propagated action potential. 
The simplest way of initiating a propagating action potential is the use of a transmembrane stimulus, :math:`I_{\mathrm{tr}}`.
According to the definition in openCARP supplying a positive :math:`I_{\mathrm{tr}}` depolarizes the tissue,
a negative :math:`I_{\mathrm{tr}}` hyperpolarizes the tissue.


Experimental setup
==================

To introduce definition and usage of transmembrane stimulus currents in openCARP a simple experiment 
using a thin monolayer preparation was defined. The following input parameters are available to steer the experiment:

.. code-block:: bash

  --duration    
                Duration of simulation (ms)

  --S1-strength 
                pick transmembrane current stimulus strength in
                [uA/cm^2] = [pA/pF] considering fixed Cm (default: 20.)

  --S1-dur     
                pick transmembrane current stimulus duration in [ms


.. _tutorial-subthreshold-stimuation:

Modeling subthreshold conditions
--------------------------------

.. _exp01-ep-ti-basic:

**Experiment exp01 (subthreshold depolarizing stimulus)**
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

We start with a longer stimulus lasting for 15 ms and gradually increase the stimulus strength
until we achieve capture. This is equivalent to measuring one data point of a `strength-duration curve <https://en.wikipedia.org/wiki/Rheobase>`_
to determine chronaxie and rheobase of our preparation.

.. code-block:: bash

   ./run.py --duration 20 --S1-strength 20. --S1-dur 15 --visualize

This weak current slowly depolarizes the membrane, but does not reach threshold.
After the end of stimulation, the membrane returns to its resting state, as shown in :numref:`fig-tutorial-subthreshold-depol-stim`.

.. _fig-tutorial-subthreshold-depol-stim:

.. figure:: /images/02_01_subthreshold_stim_response.png 
   :width: 50%
   :align: center

   Membrane response to a subthresold *depolarizing* transmembrane stimulus.

Inspect the variation in membrane depolarization as a function of distance to the stimulus location 
by varying the vertex index in meshalyzer.

To understand how the stimulus definition is passed on to the simulator we inspect the command line generated by the script.
For this sake run

.. code-block:: bash

   ./run.py --duration 20 --S1-strength 20. --S1-dur 15 --dry

The stimulus section of the command line is then

.. code-block:: bash

   -num_stim 1 \                   # use one stimulus electrode
   -stimulus[0].name S1 \
    \                              # pulse and type definition
   -stimulus[0].stimtype 0 \
   -stimulus[0].strength 250.0 \   
   -stimulus[0].duration 2.0 \
    \                              # electrode definition (block)       
   -stimulus[0].x0 -5050.0 \
   -stimulus[0].xd 100.0 \
   -stimulus[0].y0 -550.0 \
   -stimulus[0].yd 1100.0 \
   -stimulus[0].z0 -150.0 \
   -stimulus[0].zd 300.0



**Experiment exp02 (subthreshold hyperpolarizing stimulus)**
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

We repeat exp01 with a negative transmembrane stimulus current.

.. code-block:: bash

   ./run.py --duration 20 --S1-strength 20. --S1-dur 15 --visualize


As the membrane behaves essentially only passively in the subthreshold regime 
we observe the same response symmetric to the resting potential, see :numref:`fig-tutorial-subthreshold-hyper-stim`

.. _fig-tutorial-subthreshold-hyper-stim:

.. figure:: /images/02_01_subthreshold_stim_response_hyper.png 
   :width: 50%
   :align: center

   Membrane response to a subthresold *hyperpolarizing* transmembrane stimulus.


.. _tutorial-suprathreshold-stimuation:

Initiating action potential propagation
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

We repeat exp01 with a stronger transmembrane stimulus current of shorter duration.
Peak sodium current is known to be around :math:`250 \mu A/cm^2` and last for less than a millisecond.
We chose a strength of  :math:`250 \mu A/cm^2` of 1 millisecond duration to achieve suprathreshold depolarization.
Keep in mind that *suprathreshold depolarization* at the stimulus site does not automatically lead to a propagated response.
The local membrane response at the stimulus location is shown in :numref:`fig-tutorial-suprathreshold-stim`.

.. code-block:: bash

   ./run.py --duration 20 --S1-strength 100. --S1-dur 2 --visualize


.. _fig-tutorial-suprathreshold-stim:

.. figure:: /images/02_01_suprathreshold_stim_response.png 
   :width: 50%
   :align: center

   Membrane response to a suprathresold *depolarizing* transmembrane stimulus.

"""
import os
EXAMPLE_DESCRIPTIVE_NAME = 'Basics of usage '
EXAMPLE_AUTHOR = 'Gernot Plank <gernot.plank@medunigraz.at>'
EXAMPLE_DIR = os.path.dirname(__file__)
GUIinclude = True

from datetime import date
from carputils import settings
from carputils import tools
from carputils import mesh
from carputils.carpio import txt
import numpy as np

def parser():
    parser = tools.standard_parser()
    group  = parser.add_argument_group('experiment specific options')
    group.add_argument('--duration',
                        type = float,
                        default = 20.,
                        help = 'Duration of simulation in [ms] (default: 20.)')
    group.add_argument('--S1-strength',
                        type = float,
                        default = 20.,
                        help = 'pick transmembrane current stimulus strength in [uA/cm^2] = [pA/pF] considering fixed Cm (default: 20.)')
    group.add_argument('--S1-dur',
                        type = float,
                        default = 2.,
                        help = 'pick transmembrane current stimulus duration in [ms] (default: 2.)')
    return parser

def jobID(args):
    """
    Generate name of top level output directory.
    """
    today = date.today()
    return '{}_basic_{}'.format(today.isoformat(), args.duration)

@tools.carpexample(parser, jobID)
def run(args, job):

    # Generate mesh
    # Units are mm
    
    # Block which is thin in z direction
    geom = mesh.Block(size=(10., 1., 0.2))
    
    # Set fibre angle to 0, sheet angle to 0
    geom.set_fibres(0, 0, 90, 90)

    # Generate and return base name
    meshname = mesh.generate(geom)

    # Query for element labels
    _, etags, _ = txt.read(meshname + '.elem')
    etags = np.unique(etags)

    IntraTags = etags[etags != 0].tolist()
    ExtraTags = etags.tolist().copy()

    # Define the geometry of the stimulus at one end of the block
    # Units are um

    stim = ['-num_stim',             1,
            '-stimulus[0].name',    'S1',
            '-stimulus[0].stimtype', 0,
            '-stimulus[0].strength', args.S1_strength,
            '-stimulus[0].duration', args.S1_dur ]

    electrode = mesh.block_boundary_condition(geom, 'stimulus', 0, 'x', True)

    # Get basic command line, including solver options
    cmd = tools.carp_cmd(os.path.join(EXAMPLE_DIR, 'basic.par'))

    cmd += ['-simID',    job.ID,
            '-meshname', meshname,
            '-dt',       25,
            '-tend',     args.duration]
    cmd += tools.gen_physics_opts(ExtraTags=ExtraTags, IntraTags=IntraTags)
    cmd += stim + electrode

    if args.visualize:
        cmd += ['-gridout_i', 3]
        cmd += ['-gridout_e', 3]
        cmd += ['-spacedt', 0.1]

    # Run simulation 
    job.carp(cmd)

    
    # Do visualization
    if args.visualize and not settings.platform.BATCH:

        # Prepare file paths
        geom = os.path.join(job.ID, os.path.basename(meshname)+'_i')
        data = os.path.join(job.ID, 'vm.igb.gz')
        view = os.path.join(EXAMPLE_DIR, 'view_vm.mshz')

        # GUI
        if args.webGUI:
            data2 = os.path.join(job.ID, 'vm.igb')
            folderNameExp = f'/experiments/01_basic_usage_{job.ID}'
            os.mkdir(folderNameExp)
            cmdMesh = f'meshtool collect -imsh={geom} -omsh={folderNameExp}/{job.ID} -nod={data2} -ifmt=carp_txt -ofmt=ens_bin'
            outMesh = os.system(cmdMesh)
            if(outMesh == 0):
                print('Meshtool - conversion successfully')
            else:
                print('Meshtool - conversion failure')
        
        # Call meshalyzer
        else:
            job.meshalyzer(geom, data, view)



if __name__ == '__main__':
    run()    
