#!/usr/bin/env python

"""
This example introduces the concepts of region-based and gradient-based heterogeneities for
assigning spatially varying properties such as tissue conductivities and ion channel conductances. To run the experiments of this example do

.. code-block:: bash

    cd ${TUTORIALS}/02_EP_tissue/05A_Regions_vs_Gradients


Conceptual overview
===================

#) **Region-wise assignment of electrophysiological properties**

   In any particular finite element mesh, each element is part of a region defined by its tag
   (see :ref:`defining regions <region-definition>`). When cell- or tissue-scale properties are assigned 
   on a region-wise basis, the property in question is assigned homogeneously to all nodes and 
   elements that belong to that region. 

   One very important thing to note is that finite element nodes along the boundary between
   two different element regions will be assigned by openCARP to being in one region or the 
   other. The system responsible for deciding which boundary nodes belong to which regions is
   predictible but not necessarily intuitive. When using region-based tagging, it is important
   to remain mindful of the consequences of this process. Detailed information is available in 
   the example on :ref:`Cellular Dynamics Heterogeneity <cellular-dynamics-heterogeneity>`.

#) **Gradient-wise assignment of electrophysiological properties**

   As an alternative to the above-described approach, sometimes it is useful or relevant to 
   assign properties on a node-wise basis. For example, the value of an ion channel conductance
   at various points across the myocardial wall might vary linearly between two values as a 
   function of distance from the epicardium to the endocardium. This approach utilizes the
   openCARP adjustments interface (see :ref:`adjustments <adjustments>`) and is well-suited 
   for defining smooth variations in electrophysiological properties. Tagging is a poor option
   for this type of problem, since it might require a large number of tags OR different tag-wise
   division for gradients in different properties (e.g., a transmural gradient in one conductance
   and an apicobasal gradient in another).

Problem Setup
=============

This example will run two simulations using a 2D sheet model (1 cm x 1 cm) in which GKr and GKs 
values are smaller at the bottom of the sheet (y = 0) and larger at the top of the sheet (y = 1 cm).
In the first simulation, the domain will be subdivided into four regions, each one with a distinct
scalar multiplier for GKr and GKs: 

.. figure:: /images/02_05A_Regions_vs_Gradients_Fig1_RegionWiseGKr+GKs.png
    :scale: 60%
    :align: center

    Region-wise assignment of properties. Tissue is subdivided into four regions each of which is assigned
    different properties.

In the second simulation, the domain will consist of one unified region but the GKr and GKs values
will be scaled continuously (via linear gradient) from the bottom to the top:

.. figure:: /images/02_05A_Regions_vs_Gradients_Fig2_GradientWiseGKr+GKs.png
    :scale: 60%
    :align: center
  
    Gradient-wise assignment of properties. There is only one tissue region and GKr/GKs values vary
    continuously from low (bottom) to high (top).

Usage
=============

To run the experiments of this example do

.. code-block:: bash

    cd ${TUTORIALS}/02_EP_tissue/05A_Regions_vs_Gradients

The following optional arguments are available (default values are indicated):

.. code-block:: bash

    ./run.py --min_scf     = 0.1      # lower bound scaling factor (min = 0.000, max = 0.999)
             --max_scf     = 2.0      # upper bound scaling factor (min = 1.000, max = 10.00)

If the program is run with the ``--visualize`` option, meshalyzer will automatically load a map 
showing the difference in action potential durations between the two simulations:

.. code-block:: bash

    ./run.py --visualize

Interpreting Results
====================

With the default parameters, the program will perform the two simulations as described above.
When the vm.igb files are visualized, it will be apparent that the depolarisation patterns are
approximately the same but the repolarisation sequences are markedly different due to the 
variable representations of GKr/GKs heterogeneity. The side-by-side animation shown here is 
for the vm sequences produced by the test using the default parameters: 

.. _fig-region-vs-gradient-Vm-comparison:

.. figure:: /images/02_05A_Regions_vs_Gradients_Fig3_VmComparison.gif
    :scale: 100%
    :align: center

    Comparision of membrane voltage over time (:math:`V_m(t)`) for Region-wise (left) and
    Gradient-wise (right) variability in GKr/GKs.

The program will also produce APD.dat files in the relevant simulation subdirectories, which
can also be visualized in meshalyzer to appreciate the exact distributions of APD:

.. _fig-region-vs-gradient-APD-comparison:

.. figure:: /images/02_05A_Regions_vs_Gradients_Fig4_APDComparison.png
    :scale: 100%
    :align: center

    Comparision of action potential duration for Region-wise (left) and
    Gradient-wise (right) variability in GKr/GKs.

Notably, although the effect on repolarisation is definitely more pronounced than that on 
depolarisation, the resulting differences in APD are subtle, as shown by the maps that can
be seen using the --visualize function:

.. _fig-Region-vs-Gradient-DeltaAPD:

.. figure:: /images/02_05A_Regions_vs_Gradients_Fig5_DeltaAPD.png
    :scale: 100%
    :align: center

    Difference in action potential duration (APD) between region-based and gradient based 
    repolarisation heterogeneity

"""
import os
EXAMPLE_DESCRIPTIVE_NAME = 'Concept: regions vs. gradients'
EXAMPLE_AUTHOR = 'Patrick Boyle <pmjboyle@gmail.com>'
EXAMPLE_DIR = os.path.dirname(__file__)
GUIinclude = False

import sys
import numpy as np
from datetime import date

from carputils.carpio import txt
from carputils import settings
from carputils import tools
from carputils import mesh

isPy2 = True
if sys.version_info.major > 2:
    isPy2  = False
    xrange = range


def setupMesh(args):
    # user input
    UM2MM      = 1./1000
    slabsize   = args.slabsize * UM2MM                      # in mm
    resolution = args.resolution * UM2MM                    # in mm
    centre     = np.asarray([slabsize/2., slabsize/2., 0.]) # in mm

    # generate mesh (units in mm) and return basename
    # lower left of the mesh will be the origin
    geom = mesh.Block(size = (slabsize,    slabsize,   0.),
                      centre = centre,
                      etype = 'tri',
                      resolution = resolution)

    # incorporate tag regions (divide block into four regions along y-axis)
    # ^ y
    # |
    # o -> x
    #
    # ------------ (y_r3)
    # |  tag 3   |
    # ------------ (y_r2)
    # |  tag 2   |
    # ------------ (y_r1)
    # |  tag 1   |
    # ------------ (y_r0)
    # |  tag 0   |
    # ------------ (0)

    y_r0   = 1.*slabsize/4. # in mm
    y_r1   = 2.*slabsize/4. # in mm
    y_r2   = 3.*slabsize/4. # in mm
    y_r3   = 4.*slabsize/4. # in mm

    # define blocks relative to the mesh centre (mesher wants it this way):
    ll_box_r0 = np.asarray([0.,   0., 0.])-centre; ur_box_r0 = np.asarray([slabsize, y_r0, 0.])-centre # in mm
    ll_box_r1 = np.asarray([0., y_r0, 0.])-centre; ur_box_r1 = np.asarray([slabsize, y_r1, 0.])-centre # in mm
    ll_box_r2 = np.asarray([0., y_r1, 0.])-centre; ur_box_r2 = np.asarray([slabsize, y_r2, 0.])-centre # in mm
    ll_box_r3 = np.asarray([0., y_r2, 0.])-centre; ur_box_r3 = np.asarray([slabsize, y_r3, 0.])-centre # in mm

    box_r0 = mesh.BoxRegion(ll_box_r0, ur_box_r0, 0) # in mm
    box_r1 = mesh.BoxRegion(ll_box_r1, ur_box_r1, 1) # in mm
    box_r2 = mesh.BoxRegion(ll_box_r2, ur_box_r2, 2) # in mm
    box_r3 = mesh.BoxRegion(ll_box_r3, ur_box_r3, 3) # in mm

    geom.add_region(box_r0) # mesher expects boxes relative to the center!!!
    geom.add_region(box_r1)
    geom.add_region(box_r2)
    geom.add_region(box_r3)

    meshname = mesh.generate(geom)

    return meshname, geom

def parser():
    parser = tools.standard_parser()
    group  = parser.add_argument_group('experiment specific options')

    group.add_argument('--min_scf',
                        default = 0.1, type = float,
                        help = 'minimum scaling factor used in lower extremity of tissue block')
    group.add_argument('--max_scf',
                        default = 2.0, type = float,
                        help = 'maximum scaling factor used in upper extremity of tissue block')
    group.add_argument('--slabsize',
                        default = 10000., type = float,
                        help = 'Side length of 2D slab (triangles) in microns.')
    group.add_argument('--resolution',
                        default = 100., type = float,
                        help = 'Edge length of triangular slab elements in microns.')
    return parser

def jobID(args):
    """
    Generate name of top level output directory.
    """
    today = date.today()
    return '{}'.format(today.isoformat())


@tools.carpexample(parser, jobID, clean_pattern = '{}*|(mesh)'.format(date.today()))
def run(args, job):

    # user input
    f_min_scf = args.min_scf
    f_max_scf = args.max_scf

    # baseline gKr/gKs values in the UCLA_RAB model
    gKr_base  = 0.012500
    gKs_base  = 0.173558


    # generate mesh and return basename
    meshname, geom = setupMesh(args)

    # Query for element labels
    _, etags, _ = txt.read(meshname + '.elem')
    etags = np.unique(etags)
    IntraTags = etags.tolist().copy() #etags[etags != 0].tolist() # 0 is a myocardial tag! By convention this should not be the case!
    ExtraTags = etags.tolist().copy()


    # setup gKr/gKs region-wise im_param arguments
    reg0_cmd = 'gKr*{},gKs*{}'.format(f_min_scf, f_min_scf)
    reg1_cmd = 'gKr*{},gKs*{}'.format(f_min_scf+(1./3.)*(f_max_scf-f_min_scf),
                                      f_min_scf+(1./3.)*(f_max_scf-f_min_scf))
    reg2_cmd = 'gKr*{},gKs*{}'.format(f_min_scf+(2./3.)*(f_max_scf-f_min_scf),
                                      f_min_scf+(2./3.)*(f_max_scf-f_min_scf))
    reg3_cmd = 'gKr*{},gKs*{}'.format(f_max_scf, f_max_scf)


    # setup command line and run simulation #1 (Region-wise APD variability)
    simID1 = job.ID + '_RegionWise_{0:4.2f}-{1:4.2f}'.format(f_min_scf, f_max_scf)
    cmd    = tools.carp_cmd(os.path.join(EXAMPLE_DIR, 'Region-Wise-APD-Variability.par'))
    cmd   += ['-simID',                  simID1,
              '-imp_region[0].im_param', reg0_cmd,
              '-imp_region[1].im_param', reg1_cmd,
              '-imp_region[2].im_param', reg2_cmd,
              '-imp_region[3].im_param', reg3_cmd]

    # add stimulus location (left side)
    #stimID = 0
    #cmd   += mesh.block_boundary_condition(geom, 'stimulus', stimID, 'x', True)
    lstim = ['-stimulus[0].x0',  0,
             '-stimulus[0].xd',  args.resolution*5.,
             '-stimulus[0].y0',  0,
             '-stimulus[0].yd',  args.slabsize,
             '-stimulus[0].z0', -args.resolution/2.,
             '-stimulus[0].zd',  args.resolution/2.]
    # add stimulus
    cmd += lstim
    # add phys region
    cmd += tools.gen_physics_opts(ExtraTags=ExtraTags, IntraTags=IntraTags)
    # add meshname
    cmd += ['-meshname', meshname]

    # run simulation #1
    job.carp(cmd)


    # -------------------------------------------------------------------------


    # Setup .adj and .dat files for gradient by reading mesh .pts file and extracting y values
    adjf_gKr_fn = job.ID + '_GradientWise_{0:4.2f}-{1:4.2f}_gKr'.format(f_min_scf, f_max_scf)
    adjf_gKs_fn = job.ID + '_GradientWise_{0:4.2f}-{1:4.2f}_gKs'.format(f_min_scf, f_max_scf)



    # Read in .pts file and output .adj and .dat files
    points, npts = txt.read(meshname + '.pts')
    if not args.dry:
        adjf_gKr = open(adjf_gKr_fn + '.adj', 'w')
        adjf_gKs = open(adjf_gKs_fn + '.adj', 'w')
        datf_gKr = open(adjf_gKr_fn + '.dat', 'w')
        datf_gKs = open(adjf_gKs_fn + '.dat', 'w')

        for i in range(npts):
            if not i:
                # Write .adj file headers
                adjf_gKr.write('{}\nintra\n'.format(npts))
                adjf_gKs.write('{}\nintra\n'.format(npts))

            # y values will vary from 0 to 'slabsize' from bottom to top of sheet
            scf = f_min_scf + ((f_max_scf - f_min_scf) * points[i,1]/args.slabsize)

            adjf_gKr.write('{} {}\n'.format(i, scf*gKr_base))
            adjf_gKs.write('{} {}\n'.format(i, scf*gKs_base))
            datf_gKr.write('{}\n'.format(scf*gKr_base))
            datf_gKs.write('{}\n'.format(scf*gKs_base))

        adjf_gKr.close()
        adjf_gKs.close()
        datf_gKr.close()
        datf_gKs.close()


    # Setup command line and run simulation #2 (Gradient-wise APD variability)
    simID2 = job.ID + '_GradientWise_{0:4.2f}-{1:4.2f}'.format(f_min_scf, f_max_scf)
    cmd  = tools.carp_cmd(os.path.join(EXAMPLE_DIR, 'Gradient-Wise-APD-Variability.par'))
    cmd += ['-simID',              simID2,
            '-adjustment[0].file', adjf_gKr_fn + '.adj',
            '-adjustment[1].file', adjf_gKs_fn + '.adj']
    # add stimulus
    cmd += lstim
    # add phys region
    cmd += tools.gen_physics_opts(ExtraTags=ExtraTags, IntraTags=IntraTags)
    # add meshname
    cmd += ['-meshname', meshname]
    job.carp(cmd)

    if args.dry:
        return
    
    # Calculate and output difference in APD between Sim1 and Sim2 for each node in the grid
    depolf1 = txt.read(os.path.join(simID1, 'init_acts_depol-thresh.dat'))
    repolf1 = txt.read(os.path.join(simID1, 'init_acts_repol-thresh.dat'))
    APD1    = repolf1 - depolf1
    txt.write(os.path.join(simID1, 'APD.dat'), APD1)


    depolf2 = txt.read(os.path.join(simID2, 'init_acts_depol-thresh.dat'))
    repolf2 = txt.read(os.path.join(simID2, 'init_acts_repol-thresh.dat'))
    APD2    = repolf2 - depolf2
    txt.write(os.path.join(simID2, 'APD.dat'), APD2)


    dAPD    = APD2 - APD1
    dAPD_fn = job.ID + '_DeltaAPD_{0:4.2f}-{1:4.2f}.dat'.format(f_min_scf, f_max_scf)
    txt.write(dAPD_fn, dAPD)


    # Do visualization
    if args.visualize and not settings.platform.BATCH:
        geom = meshname
        data = os.path.join(dAPD_fn)
        view = os.path.join(EXAMPLE_DIR, 'DeltaAPD.mshz')

        # GUI
        if args.webGUI:
            folderNameExp = f'/experiments/05A_Regions_vs_Gradients_{job.ID}'
            os.mkdir(folderNameExp)
            cmdMesh = f'meshtool collect -imsh=/tutorials/02_EP_tissue/05A_Regions_vs_Gradients/{geom} -omsh={folderNameExp}/{job.ID} -nod=/tutorials/02_EP_tissue/05A_Regions_vs_Gradients/{data} -ifmt=carp_txt -ofmt=ens_bin'
            outMesh = os.system(cmdMesh)
            if(outMesh == 0):
                print('Meshtool - conversion successfully')
            else:
                print('Meshtool - conversion failure')
        
        # Meshalyzer
        else:
            job.meshalyzer(geom, data, view)



if __name__ == '__main__':
    run()
